from importlib.util import find_spec

from fed_rag.exceptions import MissingExtraError


def check_huggingface_evals_installed(cls_name: str | None = None) -> None:
    datasets_spec = find_spec("datasets")

    has_huggingface = datasets_spec is not None

    if not has_huggingface:
        if cls_name:
            msg = (
                f"`{cls_name}` requires the `huggingface-evals` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface-evals]`."
            )
        else:
            msg = (
                "Missing installation of the huggingface-evals extra, yet is required "
                "by an import `HuggingFaceBenchmark` class. To fix please run "
                "`pip install fed-rag[huggingface-evals]`."
            )

        raise MissingExtraError(msg)



import pytest

from fed_rag.evals import ExactMatchEvaluationMetric


@pytest.mark.parametrize(
    ("pred", "actual", "expected"),
    [
        ("1+1=2", "1+1=2", 1.0),
        ("Yes, Correct!", "yes, correct!", 1.0),
        ("not the same", "as me", 0.0),
    ],
    ids=["match", "match case insensitive", "not match"],
)
def test_exact_match(pred: str, actual: str, expected: float) -> None:
    metric = ExactMatchEvaluationMetric()

    # act
    res = metric(prediction=pred, actual=actual)

    assert res == expected



"""Base MCP Knowledge Source Class"""

from abc import ABC, abstractmethod
from typing import Any

from mcp.types import CallToolResult
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr
from typing_extensions import Self

from fed_rag.data_structures import KnowledgeNode

from .utils import CallToolResultConverter


class BaseMCPKnowledgeSource(BaseModel, ABC):
    model_config = ConfigDict(
        arbitrary_types_allowed=True, validate_assignment=True
    )
    name: str
    tool_name: str | None = None
    query_param_name: str
    tool_call_kwargs: dict[str, Any] = Field(default_factory=dict)
    _converter_fn: CallToolResultConverter = PrivateAttr()

    @abstractmethod
    async def retrieve(self, query: str) -> CallToolResult:
        """Asynchronously retrieve knowledge from this source."""

    # Common methods all sources share
    def call_tool_result_to_knowledge_nodes_list(
        self, result: CallToolResult
    ) -> list[KnowledgeNode]:
        return self._converter_fn(result=result, metadata=self.model_dump())

    def with_converter(self, converter_fn: CallToolResultConverter) -> Self:
        """Setter for converter_fn.

        Supports fluent pattern: `source = MCPStreamableHttpKnowledgeSource(...).with_converter()`
        """
        self._converter_fn = converter_fn
        return self

    def with_name(self, name: str) -> Self:
        """Setter for name.

        For convenience and users who prefer the fluent style.
        """

        self.name = name
        return self

    def with_query_param_name(self, v: str) -> Self:
        """Setter for query param name.

        For convenience and users who prefer the fluent style.
        """

        self.query_param_name = v
        return self

    def with_tool_call_kwargs(self, v: dict[str, Any]) -> Self:
        """Setter for tool call kwargs.

        For convenience and users who prefer the fluent style.
        """

        self.tool_call_kwargs = v
        return self



"""Base Benchmark and Benchmarker"""

from abc import ABC, abstractmethod
from typing import Any, Generator, Iterator, Sequence

from pydantic import BaseModel, ConfigDict, PrivateAttr, model_validator

from fed_rag.data_structures.evals import BenchmarkExample
from fed_rag.exceptions import BenchmarkGetExamplesError, BenchmarkParseError


class BaseBenchmark(BaseModel, ABC):
    """Base Benchmark."""

    _examples: Sequence[BenchmarkExample] = PrivateAttr()

    model_config = ConfigDict(arbitrary_types_allowed=True)

    # give it a sequence interface for accessing examples more easily
    def __getitem__(self, index: int) -> BenchmarkExample:
        return self._examples.__getitem__(index)

    def __len__(self) -> int:
        return self._examples.__len__()

    # shouldn't override Pydantic BaseModels' __iter__
    def as_iterator(self) -> Iterator[BenchmarkExample]:
        return self._examples.__iter__()

    @model_validator(mode="after")
    def set_examples(self) -> "BaseBenchmark":
        try:
            self._examples = self._get_examples()
        except BenchmarkParseError as e:
            raise BenchmarkGetExamplesError(
                f"Failed to parse examples: {str(e)}"
            ) from e
        except Exception as e:
            raise (
                BenchmarkGetExamplesError(f"Failed to get examples: {str(e)}")
            ) from e
        return self

    # abstractmethods
    @abstractmethod
    def _get_examples(self, **kwargs: Any) -> Sequence[BenchmarkExample]:
        """Method to get examples."""

    @abstractmethod
    def as_stream(self) -> Generator[BenchmarkExample, None, None]:
        """Produce a stream of `BenchmarkExamples`."""

    @property
    @abstractmethod
    def num_examples(self) -> int:
        """Number of examples in the benchmark.

        NOTE: if streaming, `_examples` is likely set to an empty list. Thus,
        we leave this implementation for the subclasses.
        """



"""Public RAG Trainer Managers API"""

from .huggingface import HuggingFaceRAGTrainerManager
from .pytorch import PyTorchRAGTrainerManager

__all__ = ["HuggingFaceRAGTrainerManager", "PyTorchRAGTrainerManager"]



import re
from unittest.mock import patch

import pytest

from fed_rag.exceptions import MissingExtraError
from fed_rag.generators.huggingface.utils import check_huggingface_installed


def test_check_raises_error() -> None:
    """Check raises error from utils."""

    modules = {"transformers": None}

    with patch.dict("sys.modules", modules):
        msg = (
            "Missing installation of the huggingface extra, yet is required "
            "by an imported class. To fix please run `pip install fed-rag[huggingface]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            check_huggingface_installed()



"""MCP Knowledge Source with stdio Transport"""

import uuid
from typing import Any

from mcp import ClientSession, StdioServerParameters
from mcp.client.stdio import stdio_client
from mcp.types import CallToolResult
from typing_extensions import Self

from .base import BaseMCPKnowledgeSource
from .utils import CallToolResultConverter, default_converter


class MCPStdioKnowledgeSource(BaseMCPKnowledgeSource):
    """The MCPStdioKnowledgeSource class.

    Users can easily connect MCP tools as their source of knowledge in RAG systems
    via the stdio transport.
    """

    server_params: StdioServerParameters

    def __init__(
        self,
        server_params: StdioServerParameters,
        tool_name: str,
        query_param_name: str,
        tool_call_kwargs: dict[str, Any] | None = None,
        name: str | None = None,
        converter_fn: CallToolResultConverter | None = None,
    ):
        name = name or f"source-stdio-{str(uuid.uuid4())}"
        tool_call_kwargs = tool_call_kwargs or {}
        super().__init__(
            name=name,
            server_params=server_params,
            tool_name=tool_name,
            query_param_name=query_param_name,
            tool_call_kwargs=tool_call_kwargs,
        )
        self._converter_fn = converter_fn or default_converter

    def with_server_params(self, server_params: StdioServerParameters) -> Self:
        """Setter for server params.

        For convenience and users who prefer the fluent style.
        """

        self.server_params = server_params
        return self

    async def retrieve(self, query: str) -> CallToolResult:
        async with stdio_client(self.server_params) as (read, write):
            async with ClientSession(read, write) as session:
                # Initialize the connection
                await session.initialize()
                # Call a tool
                tool_arguments = {
                    self.query_param_name: query,
                    **self.tool_call_kwargs,
                }
                tool_result = await session.call_tool(
                    self.tool_name, arguments=tool_arguments
                )
        return tool_result



"""HotpotQA benchmark"""

from typing import Any

from pydantic import model_validator

from fed_rag.base.evals.benchmark import BaseBenchmark

from .mixin import HuggingFaceBenchmarkMixin
from .utils import check_huggingface_evals_installed


class HuggingFaceHotpotQA(HuggingFaceBenchmarkMixin, BaseBenchmark):
    """HuggingFace HotpotQA Benchmark.

    HotpotQA is a multi-hop question answering dataset that requires reasoning
    over multiple paragraphs to answer questions.

    Example schema:
        {
            "id": "5a7a06935542990198eaf050",
            "question": "Which magazine was started first Arthur's Magazine or First for Women?",
            "answer": "Arthur's Magazine",
            "type": "comparison",
            "level": "medium",
            "supporting_facts": {
                "title": ["Arthur's Magazine", "First for Women"],
                "sent_id": [0, 0]
            },
            "context": {
                "title": [
                    "Radio City (Indian radio station)",
                    "History of Albanian football",
                    "Echosmith",
                    "Women's colleges in the Southern United States",
                    "First Arthur County Courthouse and Jail",
                    "Arthur's Magazine",
                    "2012–13 Ukrainian Hockey Championship",
                    "First for Women",
                    "Freeway Complex Fire",
                    "William Rast"
                ],
                "sentences": [
                    ["Radio City is India's first private FM...", "It plays Hindi..."],
                    ["Football in Albania existed before...", "The Albanian..."],
                    ...
                    ["Arthur's Magazine (1844–1846) was an American literary periodical...", "It was founded by..."],
                    ...
                    ["First for Women is a woman's magazine published by...", "The magazine was started in 1989."],
                    ...
                ]
            }
        }
    """

    dataset_name = "hotpot_qa/hotpot_qa"
    configuration_name: str = "distractor"

    def _get_query_from_example(self, example: dict[str, Any]) -> str:
        return str(example["question"])

    def _get_response_from_example(self, example: dict[str, Any]) -> str:
        return str(example["answer"])

    def _get_context_from_example(self, example: dict[str, Any]) -> str | None:
        context = example.get("context", {})
        if isinstance(context, dict):
            titles = context.get("title", [])
            sentences = context.get("sentences", [])

            # Build context by combining titles with their sentences
            context_parts = []
            for i, (title, sents) in enumerate(zip(titles, sentences)):
                if isinstance(sents, list):
                    context_parts.append(f"{title}: {' '.join(sents)}")
                else:
                    context_parts.append(f"{title}: {sents}")

            return " ".join(context_parts) if context_parts else None
        else:
            # If context is not a dict, return None
            return None

    @model_validator(mode="before")
    @classmethod
    def _validate_extra_installed(cls, data: Any) -> Any:
        """Validate that huggingface-evals dependencies are installed."""
        check_huggingface_evals_installed(cls.__name__)
        return data



from importlib.util import find_spec

from fed_rag.exceptions import MissingExtraError


def check_huggingface_installed(cls_name: str | None = None) -> None:
    transformers_spec = find_spec("transformers")
    peft_spec = find_spec("peft")
    sentence_transformers_spec = find_spec("sentence_transformers")

    has_huggingface = (
        (transformers_spec is not None)
        and (peft_spec is not None)
        and (sentence_transformers_spec is not None)
    )
    if not has_huggingface:
        if cls_name:
            msg = (
                f"`{cls_name}` requires the `huggingface` extra to be installed. "
                "To fix please run `pip install fed-rag[huggingface]`."
            )
        else:
            msg = (
                "Missing installation of the huggingface extra, yet is required "
                "by an imported class. To fix please run `pip install fed-rag[huggingface]`."
            )

        raise MissingExtraError(msg)



import uuid
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, Callable, Generator, Iterable

from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_core.language_models import BaseLLM
from langchain_core.outputs import Generation, LLMResult
from langchain_core.vectorstores import VectorStore

from fed_rag.base.generator import BaseGenerator
from fed_rag.core.rag_system._synchronous import _RAGSystem
from fed_rag.data_structures import KnowledgeNode
from fed_rag.exceptions import BridgeError

if TYPE_CHECKING:  # pragma: no cover
    from torch import Tensor


class FedRAGVectorStore(VectorStore):
    """A ~langchain_core.vectorstores.VectorStore adapter for fed_rag._RAGSystem.

    Can be converted to a ~langchain_core.vectorstores.VectorStoreRetriever
    using the `as_retriever` method.
    """

    class FedRAGEmbeddings(Embeddings):
        def __init__(
            self, encode_query: Callable[[str | list[str]], "Tensor"]
        ):
            self.encode_query = encode_query

        def embed_documents(self, texts: list[str]) -> list[list[float]]:
            return self.encode_query(texts).tolist()  # type: ignore

        def embed_query(self, text: str) -> list[float]:
            return self.encode_query(text).tolist()  # type: ignore

    def __init__(self, rag_system: "_RAGSystem"):
        """Initialize the FedRAG Vector Store."""
        super().__init__()
        self._rag_system = rag_system

    @property
    def embeddings(self) -> Embeddings:
        """Return the embeddings used by the vector store."""
        return self.FedRAGEmbeddings(
            encode_query=self._rag_system.retriever.encode_query
        )

    def delete(self, ids: list[str] | None = None, **kwargs: Any) -> bool:
        """Delete nodes from the vector store.

        Args:
            ids (list[str], optional): List of node IDs to delete. If None, no nodes are deleted.
            **kwargs (Any): Additional keyword arguments (not used).

        Returns:
            bool: True if nodes were deleted, False if no IDs were provided.
        """
        if ids is None:
            return False
        for id_ in ids:
            self._rag_system.knowledge_store.delete_node(id_)
        return True

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: list[dict[str, Any]] | None = None,
        ids: list[str] | None = None,
        **kwargs: Any,
    ) -> list[str]:
        """Add texts to the vector store.

        Args:
            texts (Iterable[str]): Texts to add.
            metadatas (list[dict[str, Any]], optional): List of metadata dictionaries for each text. Defaults to None.
            ids (list[str], optional): List of IDs for the nodes. If None, IDs will be generated automatically.
            **kwargs (Any): Additional keyword arguments (not used).

        Returns:
            list[str]: List of node IDs for the added texts.
        """
        texts = list(texts)
        if not ids:
            ids = [str(uuid.uuid4()) for _ in texts]
        if not metadatas:
            metadatas = [{} for _ in texts]
        if len(texts) != len(metadatas) or len(texts) != len(ids):
            raise ValueError(
                "The number of texts, metadatas, and ids must match."
            )
        nodes = [
            KnowledgeNode(
                node_id=id_,
                embedding=self.embeddings.embed_query(text),
                node_type="text",
                text_content=text,
                metadata=metadata,
            )
            for text, metadata, id_ in zip(texts, metadatas, ids)
        ]
        self._rag_system.knowledge_store.load_nodes(nodes)
        return ids

    def add_documents(
        self,
        documents: list[Document],
        ids: list[str] | None = None,
        **kwargs: Any,
    ) -> list[str]:
        """Add documents to the vector store.

        Args:
            documents (list[Document]): List of documents to add.
            ids (list[str], optional): List of IDs for the nodes. If None, IDs will be generated automatically.
            **kwargs (Any): Additional keyword arguments (not used).

        Returns:
            list[str]: List of node IDs for the added documents.
        """
        if not ids:
            # Documents may have ids, see ~langchain_core.documents.base.BaseMedia
            ids = [doc.id or str(uuid.uuid4()) for doc in documents]
        nodes = [
            KnowledgeNode(
                node_id=id_,
                embedding=self.embeddings.embed_query(doc.page_content),
                node_type="text",
                text_content=doc.page_content,
                metadata=doc.metadata,
            )
            for doc, id_ in zip(documents, ids)
        ]
        self._rag_system.knowledge_store.load_nodes(nodes)
        return ids

    def similarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[Document]:
        """Perform similarity search.

        Args:
            query (str): The query string to search against.
            k (int): The number of top results to return. Defaults to 4.
            **kwargs (Any): Additional keyword arguments (not used).

        Returns:
            list[Document]: A list of Documents that are similar to the query.
        """
        embedding = self.embeddings.embed_query(query)
        return self.similarity_search_by_vector(embedding, k, **kwargs)

    def similarity_search_by_vector(
        self, embedding: list[float], k: int = 4, **kwargs: Any
    ) -> list[Document]:
        """Perform similarity search by vector.

        Args:
            embedding (list[float]): The embedding vector to search against.
            k (int): The number of top results to return. Defaults to 4.
            **kwargs (Any): Additional keyword arguments (not used).

        Returns:
            list[Document]: A list of Documents that are similar to the embedding.
        """
        return [
            doc
            for doc, _ in self.similarity_search_with_score_by_vector(
                embedding, k, **kwargs
            )
        ]

    def similarity_search_with_score(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> list[tuple[Document, float]]:
        """Perform similarity search with score.

        Args:
            query (str): The query string to search against.
            k (int): The number of top results to return. Defaults to 4.
            **kwargs (Any): Additional keyword arguments (not used).

        Returns:
            list[tuple[Document, float]]: A list of tuples containing the Document and its score.
        """
        embedding = self.embeddings.embed_query(query)
        return self.similarity_search_with_score_by_vector(
            embedding, k, **kwargs
        )

    def similarity_search_with_score_by_vector(
        self,
        embedding: list[float],
        k: int = 4,
        **kwargs: Any,
    ) -> list[tuple[Document, float]]:
        """Perform similarity search with score by vector.

        Args:
            embedding (list[float]): The embedding vector to search against.
            k (int): The number of top results to return. Defaults to 4.
            **kwargs (Any): Additional keyword arguments (not used).

        Returns:
            list[tuple[Document, float]]: A list of tuples containing the Document and its score.
        """
        return [
            (
                Document(
                    id=node.node_id,
                    page_content=node.text_content or "",
                    metadata=node.metadata,
                ),
                score,
            )
            for score, node in self._rag_system.knowledge_store.retrieve(
                query_emb=embedding, top_k=k
            )
        ]

    def _select_relevance_score_fn(self) -> Callable[[float], float]:
        """Select the relevance score function for the vector store.

        NOTE: Only cosine similarity is supported for now.
        """
        return self._cosine_relevance_score_fn  # type: ignore

    @classmethod
    def from_texts(cls: type[VectorStore], **kwargs: Any) -> type[VectorStore]:
        raise BridgeError(
            "FedRAGVectorStore does not support from_texts method. "
            "Use the add_texts method to add texts to the vector store."
        )


class FedRAGLLM(BaseLLM):
    def __init__(self, rag_system: "_RAGSystem"):
        """Initialize the FedRAG LLM."""
        super().__init__()
        self._rag_system = rag_system

    @property
    def _llm_type(self) -> str:
        """Return the type of the LLM."""
        return "fed_rag.generator"

    @contextmanager
    def _generator(
        self,
    ) -> Generator[BaseGenerator, None, None]:
        original_template = self._rag_system.generator.prompt_template
        try:
            template_for_llamaindex = "{query}"
            self._rag_system.generator.prompt_template = (
                template_for_llamaindex
            )

            yield self._rag_system.generator
        finally:
            self._rag_system.generator.prompt_template = original_template

    def _generate(
        self,
        prompts: list[str],
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> LLMResult:
        """Generate text using the FedRAG system.

        Args:
            prompts (list[str]): List of prompts to generate text for.
            stop (list[str], optional): List of stop sequences. Not supported in this implementation.
            **kwargs (Any): Additional keyword arguments (not used).

        Returns:
            LLMResult: The result of the text generation containing generations.
        """
        if stop is not None:
            raise BridgeError(
                "FedRAGLLM does not support stop sequences. "
                "Please use the generator directly if you need this feature."
            )
        with self._generator() as generator:
            generations = [
                [Generation(text=generator.generate(query=prompt, context=""))]
                for prompt in prompts
            ]
        return LLMResult(generations=generations)

    def _stream(
        self,
        prompt: str,
        stop: list[str] | None = None,
        run_manager: CallbackManagerForLLMRun | None = None,
        **kwargs: Any,
    ) -> Iterable[Generation]:
        """Stream text generation.

        NOTE: This method is not supported in FedRAGLLM.
        """
        raise BridgeError(
            "FedRAGLLM does not support streaming. "
            "Please use the generator directly if you need this feature."
        )



"""Base Bridge"""

import importlib.metadata
from typing import ClassVar, Optional

from packaging.version import Version
from pydantic import BaseModel, ConfigDict

from fed_rag.data_structures import BridgeMetadata, CompatibleVersions
from fed_rag.exceptions import IncompatibleVersionError, MissingExtraError


class BridgeRegistryMixin:
    """Mixin to manage bridge registration."""

    bridges: ClassVar[dict[str, BridgeMetadata]] = {}

    @classmethod
    def _register_bridge(cls, metadata: BridgeMetadata) -> None:
        """Register a bridge's metadata."""
        if metadata["framework"] not in cls.bridges:
            cls.bridges[metadata["framework"]] = metadata

    def __init_subclass__(cls) -> None:
        """Register bridges on subclass creation."""
        super().__init_subclass__()
        for base in cls.__mro__:
            if issubclass(base, BaseBridgeMixin) and hasattr(
                base, "_method_name"
            ):
                metadata = base.get_bridge_metadata()
                if hasattr(base, metadata["method_name"]):
                    cls._register_bridge(metadata)


class BaseBridgeMixin(BaseModel):
    """Base Bridge Class."""

    # Version of the bridge implementation.
    _bridge_version: ClassVar[str]
    _bridge_extra: ClassVar[Optional[str | None]]
    _framework: ClassVar[str]
    _compatible_versions: ClassVar[CompatibleVersions]
    _method_name: ClassVar[str]

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @classmethod
    def get_bridge_metadata(cls) -> BridgeMetadata:
        metadata: BridgeMetadata = {
            "bridge_version": cls._bridge_version,
            "framework": cls._framework,
            "compatible_versions": cls._compatible_versions,
            "method_name": cls._method_name,
        }
        return metadata

    @classmethod
    def _validate_framework_installed(cls) -> None:
        """Check if the framework is installed and compatible."""
        try:
            installed_version = Version(
                importlib.metadata.version(cls._framework.replace("-", "_"))
            )
            cls._check_bound("min", installed_version)
            cls._check_bound("max", installed_version)
        except importlib.metadata.PackageNotFoundError:
            missing_package_or_extra = (
                f"fed-rag[{cls._bridge_extra}]"
                if cls._bridge_extra
                else cls._framework
            )
            msg = (
                f"`{cls._framework}` module is missing but needs to be installed. "
                f"To fix please run `pip install {missing_package_or_extra}`."
            )
            raise MissingExtraError(msg)

    @classmethod
    def _check_bound(
        cls,
        bound_name: str,
        installed: Version,
    ) -> None:
        """Check one side of compatibility and raise if it fails."""
        if bound_name not in cls._compatible_versions:
            return

        req = Version(cls._compatible_versions[bound_name])
        if (bound_name == "min" and installed < req) or (
            bound_name == "max" and installed > req
        ):
            direction = "Minimum" if bound_name == "min" else "Maximum"
            raise IncompatibleVersionError(
                f"`{cls._framework}` version {installed} is incompatible. "
                f"{direction} required is {req}."
            )



from .core import FedRAGError


class TrainerError(FedRAGError):
    """Base errors for all rag trainer relevant exceptions."""

    pass


class InconsistentDatasetError(TrainerError):
    """Raised if underlying datasets between dataloaders are inconsistent."""

    pass


class InvalidLossError(TrainerError):
    """Raised if an unexpected loss is attached to a trainer object."""

    pass


class InvalidDataCollatorError(TrainerError):
    """Raised if an invalid data collator is attached to a trainer object."""

    pass


class MissingInputTensor(TrainerError):
    """Raised if a required tensor has not been supplied in the inputs."""

    pass



"""Internal RAG System Module"""

from typing import TYPE_CHECKING

from pydantic import BaseModel, ConfigDict

from fed_rag.base.bridge import BridgeRegistryMixin
from fed_rag.data_structures import RAGConfig, RAGResponse, SourceNode
from fed_rag.exceptions import RAGSystemError

if TYPE_CHECKING:  # pragma: no cover
    # to avoid circular imports, using forward refs
    from fed_rag.base.generator import BaseGenerator
    from fed_rag.base.no_encode_knowledge_store import (
        BaseNoEncodeKnowledgeStore,
    )


class _NoEncodeRAGSystem(BridgeRegistryMixin, BaseModel):
    """Unbridged implementation of NoEncodeRAGSystem.

    IMPORTANT: This is an internal implementation class.
    It should only be used by bridge mixins and never referenced directly
    by user code or other parts of the library.

    All interaction with RAG systems should be through the public NoEncodeRAGSystem
    class.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    generator: "BaseGenerator"
    knowledge_store: "BaseNoEncodeKnowledgeStore"
    rag_config: RAGConfig

    def query(self, query: str) -> RAGResponse:
        """Query the RAG system."""
        source_nodes = self.retrieve(query)
        context = self._format_context(source_nodes)
        response = self.generate(query=query, context=context)
        return RAGResponse(source_nodes=source_nodes, response=response)

    def batch_query(self, queries: list[str]) -> list[RAGResponse]:
        """Batch query the RAG system."""
        source_nodes_list = self.batch_retrieve(queries)
        contexts = [
            self._format_context(source_nodes)
            for source_nodes in source_nodes_list
        ]
        responses = self.batch_generate(queries, contexts)
        return [
            RAGResponse(source_nodes=source_nodes, response=response)
            for source_nodes, response in zip(source_nodes_list, responses)
        ]

    def retrieve(self, query: str) -> list[SourceNode]:
        """Retrieve from NoEncodeKnowledgeStore."""
        raw_retrieval_result = self.knowledge_store.retrieve(
            query=query, top_k=self.rag_config.top_k
        )
        return [
            SourceNode(score=el[0], node=el[1]) for el in raw_retrieval_result
        ]

    def batch_retrieve(self, queries: list[str]) -> list[list[SourceNode]]:
        """Batch retrieve from NoEncodeKnowledgeStore."""
        try:
            raw_retrieval_results = self.knowledge_store.batch_retrieve(
                queries=queries, top_k=self.rag_config.top_k
            )
        except NotImplementedError:
            raw_retrieval_results = [
                self.knowledge_store.retrieve(
                    query=query, top_k=self.rag_config.top_k
                )
                for query in queries
            ]
        return [
            [SourceNode(score=el[0], node=el[1]) for el in raw_result]
            for raw_result in raw_retrieval_results
        ]

    def generate(self, query: str, context: str) -> str:
        """Generate response to query with context."""
        return self.generator.generate(query=query, context=context)  # type: ignore

    def batch_generate(
        self, queries: list[str], contexts: list[str]
    ) -> list[str]:
        """Batch generate responses to queries with contexts."""
        if len(queries) != len(contexts):
            raise RAGSystemError(
                "Queries and contexts must have the same length for batch generation."
            )
        return self.generator.generate(query=queries, context=contexts)  # type: ignore

    def _format_context(self, source_nodes: list[SourceNode]) -> str:
        """Format the context from the source nodes."""
        # TODO: how to format image context
        return str(
            self.rag_config.context_separator.join(
                [node.get_content()["text_content"] for node in source_nodes]
            )
        )


def _resolve_forward_refs() -> None:
    """Resolve forward references in _RAGSystem."""

    # These imports are needed for Pydantic to resolve forward references
    # ruff: noqa: F401
    from fed_rag.base.generator import BaseGenerator
    from fed_rag.base.no_encode_knowledge_store import (
        BaseNoEncodeKnowledgeStore,
    )

    # Update forward references
    _NoEncodeRAGSystem.model_rebuild()


_resolve_forward_refs()



"""Tests for SQuAD 2.0 benchmark"""

import re
import sys
from unittest.mock import MagicMock, patch

import pytest
from datasets import Dataset

import fed_rag.evals.benchmarks as benchmarks
from fed_rag.data_structures.evals import BenchmarkExample
from fed_rag.exceptions import MissingExtraError


@pytest.fixture
def dummy_squad_v2_answerable() -> Dataset:
    """Create a dummy SQuAD 2.0 dataset with answerable question."""
    return Dataset.from_dict(
        {
            "id": ["56ddde2d66d3e219004dad4d"],
            "title": ["Symbiosis"],
            "context": [
                "Symbiotic relationships include those associations in which one organism lives on another (ectosymbiosis, such as..."
            ],
            "question": ["What is an example of ectosymbiosis?"],
            "answers": [{"text": ["mistletoe"], "answer_start": [114]}],
        }
    )


@pytest.fixture
def dummy_squad_v2_unanswerable() -> Dataset:
    """Create a dummy SQuAD 2.0 dataset with unanswerable question."""
    return Dataset.from_dict(
        {
            "id": ["5ad3ed26604f3c001a3ff799"],
            "title": ["Normans"],
            "context": [
                "The Normans were the people who gave their name to Normandy, a region in France."
            ],
            "question": ["What is the population of Normandy?"],
            "answers": [{"text": [], "answer_start": []}],
        }
    )


@patch("datasets.load_dataset")
def test_squad_v2_benchmark(mock_load_dataset: MagicMock) -> None:
    # arrange & act
    squad = benchmarks.HuggingFaceSQuADv2()

    # assert
    mock_load_dataset.assert_called_once_with(
        benchmarks.HuggingFaceSQuADv2.dataset_name,
        name=squad.configuration_name,
        split=squad.split,
        streaming=squad.streaming,
    )


@patch("datasets.load_dataset")
def test_squad_v2_answerable_question(
    mock_load_dataset: MagicMock, dummy_squad_v2_answerable: Dataset
) -> None:
    # arrange
    mock_load_dataset.return_value = dummy_squad_v2_answerable
    squad = benchmarks.HuggingFaceSQuADv2()

    # assert
    assert isinstance(squad[0], BenchmarkExample)
    assert squad[0].query == "What is an example of ectosymbiosis?"
    assert squad[0].response == "mistletoe"
    assert (
        squad[0].context
        == "Symbiotic relationships include those associations in which one organism lives on another (ectosymbiosis, such as..."
    )


@patch("datasets.load_dataset")
def test_squad_v2_unanswerable_question(
    mock_load_dataset: MagicMock, dummy_squad_v2_unanswerable: Dataset
) -> None:
    # arrange
    mock_load_dataset.return_value = dummy_squad_v2_unanswerable
    squad = benchmarks.HuggingFaceSQuADv2()

    # assert
    assert squad[0].response == "[NO ANSWER]"


@patch("datasets.load_dataset")
def test_squad_v2_missing_answers_field(mock_load_dataset: MagicMock) -> None:
    """Test handling when answers field is missing."""
    dataset = Dataset.from_dict(
        {
            "id": ["1"],
            "title": ["Test"],
            "context": ["Test context"],
            "question": ["Test question?"]
            # No answers field
        }
    )
    mock_load_dataset.return_value = dataset
    squad = benchmarks.HuggingFaceSQuADv2()
    assert squad[0].response == "[NO ANSWER]"


def test_huggingface_evals_extra_missing() -> None:
    """Test that proper error is raised when huggingface-evals extra is missing."""
    modules = {
        "datasets": None,
    }
    module_to_import = "fed_rag.evals.benchmarks"
    original_module = sys.modules.pop(module_to_import, None)

    with patch.dict("sys.modules", modules):
        msg = (
            "`HuggingFaceSQuADv2` requires the `huggingface-evals` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface-evals]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            import fed_rag.evals.benchmarks as benchmarks

            benchmarks.HuggingFaceSQuADv2()

    # restore module so to not affect other tests
    if original_module:
        sys.modules[module_to_import] = original_module



"""Exceptions for Bridges."""

from .core import FedRAGError


class BridgeError(FedRAGError):
    """Base bridge error for all bridge-related exceptions."""

    pass


class MissingSpecifiedConversionMethod(BridgeError):
    """Raised when bridge is missing its specified method."""

    pass


class IncompatibleVersionError(FedRAGError):
    """Raised when a fed-rag component is not compatible with the current version."""



"""Exceptions for Retrievers."""

from .core import FedRAGError, FedRAGWarning


class RetrieverError(FedRAGError):
    """Base evals error for all retriever-related exceptions."""

    pass


class RetrieverWarning(FedRAGWarning):
    """Base inspector warning for all retriever-related warnings."""

    pass



import sys

import pytest

DEPRECATED_IMPORTS = [
    ("fed_rag.types.bridge", "BridgeMetadata"),
    ("fed_rag.types.results", "TrainResult"),
    ("fed_rag.types.results", "TestResult"),
    ("fed_rag.types.knowledge_node", "KnowledgeNode"),
    ("fed_rag.types.knowledge_node", "NodeType"),
    ("fed_rag.types.knowledge_node", "NodeContent"),
    ("fed_rag.types.rag", "RAGConfig"),
    ("fed_rag.types.rag", "RAGResponse"),
    ("fed_rag.types.rag", "SourceNode"),
]


@pytest.mark.parametrize("module_path,class_name", DEPRECATED_IMPORTS)
def test_import_from_types_raises_deprecation_warning(
    module_path: str, class_name: str
) -> None:
    """Test that importing from deprecated types modules raises warnings."""

    # clear the module from sys.modules if it exists
    if module_path in sys.modules:
        del sys.modules[module_path]

    with pytest.warns(DeprecationWarning):
        import importlib

        module = importlib.import_module(module_path)
        getattr(module, class_name)  # ensure its loaded



"""Tests for BoolQ benchmark"""

from unittest.mock import MagicMock, patch

import pytest
from datasets import Dataset

import fed_rag.evals.benchmarks as benchmarks
from fed_rag.data_structures.evals import BenchmarkExample


@pytest.fixture
def dummy_boolq() -> Dataset:
    """Create a dummy BoolQ dataset for testing."""
    return Dataset.from_dict(
        {
            "question": ["is confectionary sugar the same as powdered sugar"],
            "answer": [True],
            "passage": [
                "Powdered sugar, also called confectioners' sugar, is a finely ground sugar ..."
            ],
        }
    )


@patch("datasets.load_dataset")
def test_boolq_query_response_context_extractors(
    mock_load_dataset: MagicMock, dummy_boolq: Dataset
) -> None:
    mock_load_dataset.return_value = dummy_boolq
    boolq = benchmarks.HuggingFaceBoolQ()

    assert isinstance(boolq[0], BenchmarkExample)
    assert (
        boolq[0].query == "is confectionary sugar the same as powdered sugar"
    )
    assert boolq[0].response == "true"
    assert (
        boolq[0].context
        == "Powdered sugar, also called confectioners' sugar, is a finely ground sugar ..."
    )


@patch("datasets.load_dataset")
def test_boolq_false_response(mock_load_dataset: MagicMock) -> None:
    dataset = Dataset.from_dict(
        {
            "question": ["is elder scrolls online the same as skyrim"],
            "answer": [False],
            "passage": [
                "As with other games in The Elder Scrolls series, the game is set on the continent of Tamriel."
            ],
        }
    )
    mock_load_dataset.return_value = dataset
    boolq = benchmarks.HuggingFaceBoolQ()

    assert boolq[0].response == "false"



"""Bridge type definitions for fed-rag."""

from typing import TypedDict


class CompatibleVersions(TypedDict, total=False):
    """Type definition for compatible versions.

    Defines optional, inclusive version bounds for compatibility checks.
    """

    min: str
    max: str


class BridgeMetadata(TypedDict):
    """Type definition for bridge metadata."""

    bridge_version: str
    framework: str
    compatible_versions: CompatibleVersions
    method_name: str



"""Data structures for fed_rag.evals"""

from enum import Enum

from pydantic import BaseModel

from .rag import RAGResponse


class BenchmarkExample(BaseModel):
    """Benchmark example data class."""

    query: str
    response: str
    context: str | None = None


class BenchmarkResult(BaseModel):
    """Benchmark result data class."""

    score: float
    metric_name: str
    num_examples_used: int
    num_total_examples: int
    evaluations_file: str | None


class BenchmarkEvaluatedExample(BaseModel):
    """Evaluated benchmark example data class."""

    score: float
    example: BenchmarkExample
    rag_response: RAGResponse

    def model_dump_json_without_embeddings(self) -> str:
        return self.model_dump_json(
            exclude={
                "rag_response": {
                    "source_nodes": {"__all__": {"node": {"embedding"}}}
                }
            }
        )


class AggregationMode(str, Enum):
    """Mode for aggregating evaluation scores."""

    AVG = "avg"
    SUM = "sum"
    MAX = "max"
    MIN = "min"



"""Internal Async RAG System Module"""

import asyncio
from typing import TYPE_CHECKING

from pydantic import BaseModel, ConfigDict

from fed_rag.base.bridge import BridgeRegistryMixin
from fed_rag.data_structures import RAGConfig, RAGResponse, SourceNode
from fed_rag.exceptions import RAGSystemError

if TYPE_CHECKING:  # pragma: no cover
    # to avoid circular imports, using forward refs
    from fed_rag.base.generator import BaseGenerator
    from fed_rag.base.knowledge_store import BaseAsyncKnowledgeStore
    from fed_rag.base.retriever import BaseRetriever


class _AsyncRAGSystem(BridgeRegistryMixin, BaseModel):
    """Unbridged implementation of AsyncRAGSystem.

    IMPORTANT: This is an internal implementation class.
    It should only be used by bridge mixins and never referenced directly
    by user code or other parts of the library.

    All interaction with RAG systems should be through the public AsyncRAGSystem class.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    generator: "BaseGenerator"
    retriever: "BaseRetriever"
    knowledge_store: "BaseAsyncKnowledgeStore"
    rag_config: RAGConfig

    async def query(self, query: str) -> RAGResponse:
        """Query the RAG system."""
        source_nodes = await self.retrieve(query)
        context = self._format_context(source_nodes)
        response = await self.generate(query=query, context=context)
        return RAGResponse(source_nodes=source_nodes, response=response)

    async def batch_query(self, queries: list[str]) -> list[RAGResponse]:
        """Batch query the RAG system."""
        source_nodes_list = await self.batch_retrieve(queries)
        contexts = [
            self._format_context(source_nodes)
            for source_nodes in source_nodes_list
        ]
        responses = await self.batch_generate(queries, contexts)
        return [
            RAGResponse(source_nodes=source_nodes, response=response)
            for source_nodes, response in zip(source_nodes_list, responses)
        ]

    async def retrieve(self, query: str) -> list[SourceNode]:
        """Retrieve from KnowledgeStore."""
        query_emb: list[float] = self.retriever.encode_query(query).tolist()
        raw_retrieval_result = await self.knowledge_store.retrieve(
            query_emb=query_emb, top_k=self.rag_config.top_k
        )
        return [
            SourceNode(score=el[0], node=el[1]) for el in raw_retrieval_result
        ]

    async def batch_retrieve(
        self, queries: list[str]
    ) -> list[list[SourceNode]]:
        """Batch retrieve from KnowledgeStore."""
        query_embs: list[list[float]] = self.retriever.encode_query(
            queries
        ).tolist()
        try:
            raw_retrieval_results = await self.knowledge_store.batch_retrieve(
                query_embs=query_embs, top_k=self.rag_config.top_k
            )
        except NotImplementedError:
            raw_retrieval_tasks = [
                self.knowledge_store.retrieve(
                    query_emb=query_emb, top_k=self.rag_config.top_k
                )
                for query_emb in query_embs
            ]
            raw_retrieval_results = await asyncio.gather(*raw_retrieval_tasks)

        return [
            [SourceNode(score=el[0], node=el[1]) for el in raw_result]
            for raw_result in raw_retrieval_results
        ]

    async def generate(self, query: str, context: str) -> str:
        """Generate response to query with context."""
        return self.generator.generate(query=query, context=context)  # type: ignore

    async def batch_generate(
        self, queries: list[str], contexts: list[str]
    ) -> list[str]:
        """Batch generate responses to queries with contexts."""
        if len(queries) != len(contexts):
            raise RAGSystemError(
                "Queries and contexts must have the same length for batch generation."
            )
        return self.generator.generate(query=queries, context=contexts)  # type: ignore

    def _format_context(self, source_nodes: list[SourceNode]) -> str:
        """Format the context from the source nodes."""
        # TODO: how to format image context
        return str(
            self.rag_config.context_separator.join(
                [node.get_content()["text_content"] for node in source_nodes]
            )
        )


def _resolve_forward_refs() -> None:
    """Resolve forward references in _RAGSystem."""

    # These imports are needed for Pydantic to resolve forward references
    # ruff: noqa: F401
    from fed_rag.base.generator import BaseGenerator
    from fed_rag.base.knowledge_store import BaseAsyncKnowledgeStore
    from fed_rag.base.retriever import BaseRetriever

    # Update forward references
    _AsyncRAGSystem.model_rebuild()


_resolve_forward_refs()



import asyncio
import re

import pytest

from fed_rag.exceptions import FedRAGError
from fed_rag.utils.asyncio import asyncio_run


async def simple_async_function(value: int = 42) -> int:
    """Simple async function for testing."""
    await asyncio.sleep(0.001)
    return value


async def async_function_with_exception() -> None:
    """Async function that raises an exception."""
    await asyncio.sleep(0.001)
    raise RuntimeError("Test exception")


def test_simple_coroutine_execution() -> None:
    """Test running a simple coroutine."""
    result = asyncio_run(simple_async_function(123))
    assert result == 123


def test_coroutine_with_default_args() -> None:
    """Test running a coroutine with default arguments."""
    result = asyncio_run(simple_async_function())
    assert result == 42


def test_existing_but_not_running_loop() -> None:
    """Test behavior with an existing but not running loop."""
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    try:
        # Loop exists but is not running
        result = asyncio_run(simple_async_function(789))
        assert result == 789
    finally:
        loop.close()


def test_nested_asyncio_run_calls() -> None:
    """Test that nested calls work correctly."""

    async def outer_async() -> int:
        # This will run in a separate thread due to nested context
        inner_result = asyncio_run(simple_async_function(111))
        return int(inner_result * 2)

    result = asyncio_run(outer_async())
    assert result == 222


def test_coroutine_exception_propagation() -> None:
    """Test that exceptions from coroutines are properly propagated."""
    msg = (
        "Unable to execute async operation in current context. "
        "This may be due to nested event loops. Consider using nest_asyncio.apply() "
        "to allow nested event loops, or use async methods directly."
    )

    with pytest.raises(FedRAGError, match=re.escape(msg)):
        asyncio_run(async_function_with_exception())



from .mcp import (
    MCPKnowledgeStore,
    MCPStdioKnowledgeSource,
    MCPStreamableHttpKnowledgeSource,
)

__all__ = [
    "MCPKnowledgeStore",
    "MCPStdioKnowledgeSource",
    "MCPStreamableHttpKnowledgeSource",
]



import re
from unittest.mock import patch

import pytest

from fed_rag.evals.benchmarks.huggingface.utils import (
    check_huggingface_evals_installed,
)
from fed_rag.exceptions import MissingExtraError


def test_check_raises_error() -> None:
    """Check raises error from utils."""

    modules = {"datasets": None}

    with patch.dict("sys.modules", modules):
        msg = (
            "Missing installation of the huggingface-evals extra, yet is required "
            "by an import `HuggingFaceBenchmark` class. To fix please run "
            "`pip install fed-rag[huggingface-evals]`."
        )

        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            check_huggingface_evals_installed()



"""Exceptions for Tokenizer."""

from .core import FedRAGError, FedRAGWarning


class TokenizerError(FedRAGError):
    """Base evals error for all tokenizer-related exceptions."""

    pass


class TokenizerWarning(FedRAGWarning):
    """Base inspector warning for all tokenizer-related warnings."""

    pass



"""Knowledge Node

Note: The KnowledgeNOde implementation has moved to fed_rag.data_structures.knowledge_node.
This module is maintained for backward compatibility.
"""

import warnings

from ..data_structures.knowledge_node import (
    KnowledgeNode,
    NodeContent,
    NodeType,
)

warnings.warn(
    "Importing KnowledgeNode, NodeContent, and NodeType from fed_rag.types.knowledge_node"
    "is deprecated and will be removed in a future release. Use "
    "fed_rag.data_structures.knowledge_node or fed_rag.data_structures instead.",
    DeprecationWarning,
    stacklevel=2,  # point to users import statement
)

__all__ = ["KnowledgeNode", "NodeContent", "NodeType"]



from typing import Any, Generator

import pytest
from datasets import (
    Dataset,
    DatasetInfo,
    IterableDataset,
    Split,
    SplitDict,
    SplitInfo,
)


@pytest.fixture
def dummy_dataset() -> Dataset:
    split_dict = SplitDict()
    split_dict.add(SplitInfo(name="test", num_examples=3))
    benchmark_info = DatasetInfo(
        dataset_name="test_benchmark",
        description="A toy RAG dataset for testing purposes",
        splits=split_dict,
    )
    benchmark = Dataset.from_dict(
        {
            "query": ["a query", "another query", "yet another query"],
            "response": [
                "reponse to a query",
                "another response to another query",
                "yet another response to yet another query",
            ],
            "context": [
                "context for a query",
                "another context for another query",
                "yet another context for yet another query",
            ],
        },
        info=benchmark_info,
        split=Split.TEST,
    )
    return benchmark


@pytest.fixture
def dummy_iterable_dataset() -> IterableDataset:
    def example_gen() -> Generator[dict[str, Any], None, None]:
        yield {
            "query": "a query",
            "response": "response to a query",
            "context": "context for a query",
        }
        yield {
            "query": "another query",
            "response": "another response to another query",
            "context": "another context for another query",
        }
        yield {
            "query": "yet another query",
            "response": "yet another response to yet another query",
            "context": "yet another context for yet another query",
        }

    benchmark = IterableDataset.from_generator(example_gen, split=Split.TEST)
    split_dict = SplitDict()
    split_dict.add(SplitInfo(name="test", num_examples=3))
    benchmark.info.update(
        DatasetInfo(
            dataset_name="test_benchmark",
            description="A toy RAG dataset for testing purposes",
            splits=split_dict,
        )
    )
    return benchmark


@pytest.fixture
def dummy_mmlu() -> Dataset:
    split_dict = SplitDict()
    split_dict.add(SplitInfo(name="test", num_examples=1))
    benchmark_info = DatasetInfo(
        dataset_name="cais/mmlu",
        splits=split_dict,
    )
    benchmark = Dataset.from_dict(
        {
            "question": [
                "What is the embryological origin of the hyoid bone?"
            ],
            "choices": [
                [
                    "The first pharyngeal arch",
                    "The first and second pharyngeal arches",
                    "The second pharyngeal arch",
                    "The second and third pharyngeal arches",
                ]
            ],
            "answer": [3],
        },
        info=benchmark_info,
        split=Split.TEST,
    )
    return benchmark



import asyncio
import concurrent.futures
from typing import Any, Coroutine

from fed_rag.exceptions import FedRAGError


def asyncio_run(coro: Coroutine) -> Any:
    """
    Safely run a coroutine in any async context.

    Gets an existing event loop to run the coroutine if available.
    If there is no existing event loop, creates a new one.
    If an event loop is already running, uses threading to run in a separate thread.

    Args:
        coro: The coroutine to execute

    Returns:
        The result of the coroutine execution

    Raises:
        FedRAGError: If unable to execute the coroutine in any context

    Note:
        Inspired by LlamaIndex's approach to handling nested event loops.
        See: https://github.com/run-llama/llama_index
    """
    try:
        # Check if there's an existing event loop
        loop = asyncio.get_event_loop()

        # Check if the loop is already running
        if loop.is_running():
            # If loop is already running, run in a separate thread
            def run_coro_in_thread() -> Any:
                new_loop = asyncio.new_event_loop()
                asyncio.set_event_loop(new_loop)
                try:
                    return new_loop.run_until_complete(coro)
                finally:
                    new_loop.close()

            with concurrent.futures.ThreadPoolExecutor(
                max_workers=1
            ) as executor:
                future = executor.submit(run_coro_in_thread)
                return future.result()
        else:
            # If we're here, there's an existing loop but it's not running
            return loop.run_until_complete(coro)

    except RuntimeError:
        # If we can't get the event loop, we're likely in a different thread
        try:
            return asyncio.run(coro)
        except RuntimeError:
            raise FedRAGError(
                "Unable to execute async operation in current context. "
                "This may be due to nested event loops. Consider using nest_asyncio.apply() "
                "to allow nested event loops, or use async methods directly."
            )



"""Data structures for results

Note: The correct module has moved to fed_rag.data_structures.results. This module is
maintained for backward compatibility.
"""

import warnings

from ..data_structures.results import TestResult, TrainResult

warnings.warn(
    "Importing TrainResult, TestResult from fed_rag.types.results"
    "is deprecated and will be removed in a future release. Use "
    "fed_rag.data_structures.results or fed_rag.data_structures instead.",
    DeprecationWarning,
    stacklevel=2,  # point to users import statement
)

__all__ = ["TrainResult", "TestResult"]



import re
from unittest.mock import patch

import pytest

from fed_rag.exceptions import MissingExtraError
from fed_rag.generators.unsloth.utils import check_unsloth_installed


def test_check_raises_error() -> None:
    """Check raises error from utils."""

    modules = {"unsloth": None}

    with patch.dict("sys.modules", modules):
        # without class name
        msg = (
            "Missing installation of the `unsloth` extra, yet is required "
            "by an imported class. To fix please run `pip install fed-rag[unsloth]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            check_unsloth_installed()

        # with class name
        msg = (
            "`FakeClass` requires the `unsloth` extra to be installed. "
            "To fix please run `pip install fed-rag[unsloth]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            check_unsloth_installed("FakeClass")



from importlib.util import find_spec

from fed_rag.exceptions import MissingExtraError


def check_unsloth_installed(cls_name: str | None = None) -> None:
    unsloth_spec = find_spec("unsloth")

    has_unsloth = unsloth_spec is not None
    if not has_unsloth:
        if cls_name:
            msg = (
                f"`{cls_name}` requires the `unsloth` extra to be installed. "
                "To fix please run `pip install fed-rag[unsloth]`."
            )
        else:
            msg = (
                "Missing installation of the `unsloth` extra, yet is required "
                "by an imported class. To fix please run `pip install fed-rag[unsloth]`."
            )

        raise MissingExtraError(msg)



from .sources import MCPStdioKnowledgeSource, MCPStreamableHttpKnowledgeSource
from .store import MCPKnowledgeStore

__all__ = [
    "MCPKnowledgeStore",
    "MCPStdioKnowledgeSource",
    "MCPStreamableHttpKnowledgeSource",
]



import re
import sys
from unittest.mock import MagicMock, patch

import pytest

from fed_rag.base.tokenizer import BaseTokenizer
from fed_rag.exceptions import MissingExtraError
from fed_rag.tokenizers.unsloth_pretrained_tokenizer import (
    UnslothPretrainedTokenizer,
)


def test_hf_pretrained_generator_class() -> None:
    names_of_base_classes = [
        b.__name__ for b in UnslothPretrainedTokenizer.__mro__
    ]
    assert BaseTokenizer.__name__ in names_of_base_classes


def test_unsloth_extra_missing() -> None:
    """Test extra is not installed."""

    modules = {"unsloth": None}
    module_to_import = "fed_rag.tokenizers.unsloth_pretrained_tokenizer"

    if module_to_import in sys.modules:
        original_module = sys.modules.pop(module_to_import)

    with patch.dict("sys.modules", modules):
        msg = (
            "`UnslothPretrainedTokenizer` requires the `unsloth` extra to be installed. "
            "To fix please run `pip install fed-rag[unsloth]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            from fed_rag.tokenizers.unsloth_pretrained_tokenizer import (
                UnslothPretrainedTokenizer,
            )

            mock_tokenizer = MagicMock()

            UnslothPretrainedTokenizer(mock_tokenizer, "fake_name")

    # restore module so to not affect other tests
    sys.modules[module_to_import] = original_module



"""Tests for HellaSwag benchmark"""

from unittest.mock import MagicMock, patch

import pytest
from datasets import Dataset

import fed_rag.evals.benchmarks as benchmarks
from fed_rag.data_structures.evals import BenchmarkExample


@pytest.fixture
def dummy_hellaswag() -> Dataset:
    """Create a dummy HellaSwag dataset for testing."""
    return Dataset.from_dict(
        {
            "ind": [4],
            "activity_label": ["Removing ice from car"],
            "ctx_a": [
                "Then, the man writes over the snow covering the window of a car, and a woman wearing winter clothes smiles."
            ],
            "ctx_b": ["then"],
            "ctx": [
                "Then, the man writes over the snow covering the window of a car, and a woman wearing winter clothes smiles. then"
            ],
            "endings": [
                [
                    ", the man adds wax to the windshield and cuts it.",
                    ", a person board a ski lift, while two men supporting the head of the person wearing winter clothes snow as the we girls sled.",
                    ", the man puts on a christmas coat, knitted with netting.",
                    ", the man continues removing the snow on his car.",
                ]
            ],
            "source_id": ["activitynet~v_-1IBHYS3L-Y"],
            "split": ["train"],
            "split_type": ["indomain"],
            "label": ["3"],
        }
    )


@patch("datasets.load_dataset")
def test_hellaswag_benchmark(
    mock_load_dataset: MagicMock, dummy_hellaswag: Dataset
) -> None:
    mock_load_dataset.return_value = dummy_hellaswag
    hellaswag = benchmarks.HuggingFaceHellaSwag()

    assert isinstance(hellaswag[0], BenchmarkExample)
    assert (
        hellaswag[0].query
        == "Then, the man writes over the snow covering the window of a car, and a woman wearing winter clothes smiles. then"
    )
    assert hellaswag[0].response == "3"
    expected_context = (
        "0: , the man adds wax to the windshield and cuts it.\n"
        "1: , a person board a ski lift, while two men supporting the head of the person wearing winter clothes snow as the we girls sled.\n"
        "2: , the man puts on a christmas coat, knitted with netting.\n"
        "3: , the man continues removing the snow on his car."
    )
    assert hellaswag[0].context == expected_context


@patch("datasets.load_dataset")
def test_hellaswag_endings_structure(
    mock_load_dataset: MagicMock, dummy_hellaswag: Dataset
) -> None:
    mock_load_dataset.return_value = dummy_hellaswag
    hellaswag = benchmarks.HuggingFaceHellaSwag()
    endings_field = hellaswag[0].context
    endings_lines = endings_field.strip().split("\n")
    assert len(endings_lines) == 4
    assert all(isinstance(line, str) and line for line in endings_lines)


@patch("datasets.load_dataset")
def test_hellaswag_label_within_range(
    mock_load_dataset: MagicMock, dummy_hellaswag: Dataset
) -> None:
    mock_load_dataset.return_value = dummy_hellaswag
    hellaswag = benchmarks.HuggingFaceHellaSwag()
    # Response should be between 0 and 3 for standard HellaSwag
    response = int(hellaswag[0].response)
    assert 0 <= response <= 3



from typing import Any

from fed_rag.base.evals.metric import BaseEvaluationMetric


class MyMetric(BaseEvaluationMetric):
    def __call__(
        self, prediction: str, actual: str, *args: Any, **kwargs: Any
    ) -> float:
        return 0.42


def test_metric_call() -> None:
    metric = MyMetric()

    score = metric("fake pred", "fake actual")

    assert score == 0.42



"""Natural Questions benchmark"""

from typing import Any

from pydantic import model_validator

from fed_rag.base.evals.benchmark import BaseBenchmark

from .mixin import HuggingFaceBenchmarkMixin
from .utils import check_huggingface_evals_installed


class HuggingFaceNaturalQuestions(HuggingFaceBenchmarkMixin, BaseBenchmark):
    """HuggingFace Natural Questions Benchmark.

    Natural Questions is a question answering dataset containing real user
    questions issued to Google search, paired with Wikipedia articles containing
    the answer. Each question may have a long answer (a paragraph) and/or a
    short answer (a span within the paragraph).

    Example schema:
        {
            "id": "5225754983651766092",
            "document": {
                "title": "Trade wind",
                "url": "https://en.wikipedia.org/wiki/Trade_wind",
                "html": "<!DOCTYPE html>...",
                "tokens": {
                    "token": ["Trade", "winds", "are", "the", "pattern", "of", "easterly", "surface", "winds", ...],
                    "is_html": [False, False, False, False, False, False, False, False, False, ...],
                    "start_byte": [0, 6, 12, 16, 20, 28, 30, 38, 46, ...],
                    "end_byte": [5, 11, 15, 19, 27, 29, 37, 45, 51, ...]
                }
            },
            "question": {
                "text": "what purpose did seasonal monsoon winds have on trade",
                "tokens": ["what", "purpose", "did", "seasonal", "monsoon", "winds", "have", "on", "trade"]
            },
            "long_answer_candidates": {
                "start_byte": [43178, 44667, 48014, 49619, 50365, 52946, 54877, 56067, 57202, 58657],
                "end_byte": [44666, 45901, 49618, 50364, 50993, 54876, 56066, 56879, 58656, 60294],
                "start_token": [44, 161, 343, 547, 628, 749, 949, 1046, 1141, 1304],
                "end_token": [161, 285, 547, 628, 701, 949, 1046, 1134, 1304, 1488],
                "top_level": [True, True, True, True, True, True, True, True, True, True]
            },
            "annotations": {
                "id": ["4323936797498927989", "13037645000009169623", "4439059471919323171", "15051359051424338858", "5332861748513580580"],
                "long_answer": [
                    {"candidate_index": -1, "start_byte": -1, "end_byte": -1, "start_token": -1, "end_token": -1},
                    {"candidate_index": -1, "start_byte": -1, "end_byte": -1, "start_token": -1, "end_token": -1},
                    {"candidate_index": -1, "start_byte": -1, "end_byte": -1, "start_token": -1, "end_token": -1},
                    {"candidate_index": 0, "start_byte": 43178, "end_byte": 44666, "start_token": 44, "end_token": 161},
                    {"candidate_index": 0, "start_byte": 43178, "end_byte": 44666, "start_token": 44, "end_token": 161}
                ],
                "short_answers": [
                    {"start_byte": [], "end_byte": [], "start_token": [], "end_token": [], "text": []},
                    {"start_byte": [], "end_byte": [], "start_token": [], "end_token": [], "text": []},
                    {"start_byte": [], "end_byte": [], "start_token": [], "end_token": [], "text": []},
                    {"start_byte": [44318], "end_byte": [44657], "start_token": [140], "end_token": [159],
                     "text": ["enabled European empire expansion into the Americas and trade routes to become established across the Atlantic and Pacific oceans"]},
                    {"start_byte": [], "end_byte": [], "start_token": [], "end_token": [], "text": []}
                ],
                "yes_no_answer": [-1, -1, -1, -1, -1]
            }
        }
    """

    dataset_name = "google-research-datasets/natural_questions"
    configuration_name: str = "default"
    split: str = (
        "validation"  # Natural Questions uses 'validation' instead of 'test'
    )

    def _get_query_from_example(self, example: dict[str, Any]) -> str:
        question = example.get("question", {})
        return str(question.get("text", ""))

    def _get_response_from_example(self, example: dict[str, Any]) -> str:
        annotations = example.get("annotations", {})

        # Get the lists from annotations
        yes_no_answers = annotations.get("yes_no_answer", [])
        short_answers_list = annotations.get("short_answers", [])
        long_answers_list = annotations.get("long_answer", [])

        if (
            not yes_no_answers
            and not short_answers_list
            and not long_answers_list
        ):
            return "[NO ANSWER]"

        # Check each annotation for valid answers
        for i, yes_no_answer in enumerate(yes_no_answers):
            # Check for yes/no answer first (-1 means no yes/no answer)
            if yes_no_answer == 1:
                return "YES"
            elif yes_no_answer == 0:
                return "NO"

            # Check for short answers
            if i < len(short_answers_list):
                short_answer = short_answers_list[i]
                texts = short_answer.get("text", [])
                if texts:  # If there are short answer texts
                    # Join multiple short answers with "and"
                    return " and ".join(str(text) for text in texts if text)

            # Check for long answers
            if i < len(long_answers_list):
                long_answer = long_answers_list[i]
                if long_answer.get("candidate_index", -1) >= 0:
                    return "[LONG ANSWER EXISTS]"

        return "[NO ANSWER]"

    def _get_context_from_example(self, example: dict[str, Any]) -> str | None:
        document = example.get("document", {})

        # Extract clean text from tokens (excluding HTML tags)
        tokens = document.get("tokens", {})
        token_list = tokens.get("token", [])
        is_html_list = tokens.get("is_html", [])

        if token_list and is_html_list:
            # Filter out HTML tokens and join text tokens
            text_tokens = [
                token
                for token, is_html in zip(token_list, is_html_list)
                if not is_html
            ]
            if text_tokens:
                return " ".join(text_tokens)

        # Fallback to title as minimal context
        title = document.get("title", "")
        return title if title else None

    @model_validator(mode="before")
    @classmethod
    def _validate_extra_installed(cls, data: Any) -> Any:
        """Validate that huggingface-evals dependencies are installed."""
        check_huggingface_evals_installed(cls.__name__)
        return data



"""LangChain Bridge"""

from typing import TYPE_CHECKING

from fed_rag._bridges.langchain._version import __version__
from fed_rag.base.bridge import BaseBridgeMixin

if TYPE_CHECKING:  # pragma: no cover
    from langchain_core.language_models import BaseLLM
    from langchain_core.vectorstores import VectorStore

    from fed_rag.core.rag_system._synchronous import (
        _RAGSystem,  # avoids circular import
    )


class LangChainBridgeMixin(BaseBridgeMixin):
    """LangChain Bridge.

    This mixin adds LangChain conversion capabilities to _RAGSystem.
    When mixed with an unbridged _RAGSystem, it allows direct conversion to
    LangChain's VectorStore and BaseLLM through the to_langchain() method.
    """

    _bridge_version = __version__
    _bridge_extra = "langchain"
    _framework = "langchain-core"
    _compatible_versions = {"min": "0.3.62"}
    _method_name = "to_langchain"

    def to_langchain(self: "_RAGSystem") -> tuple["VectorStore", "BaseLLM"]:
        """Converts the _RAGSystem to a tuple of ~langchain_core.vectorstores.VectorStore and ~langchain_core.language_models.BaseLLM."""
        self._validate_framework_installed()

        from fed_rag._bridges.langchain._bridge_classes import (
            FedRAGLLM,
            FedRAGVectorStore,
        )

        return FedRAGVectorStore(self), FedRAGLLM(self)



"""Type stubs for fed_rag module"""

# Lazy-loaded classes (type-only declarations)
# Lazy-loaded modules
from fed_rag import generators as generators
from fed_rag import retrievers as retrievers
from fed_rag import trainer_managers as trainer_managers
from fed_rag import trainers as trainers
from fed_rag.generators import HFPeftModelGenerator as HFPeftModelGenerator
from fed_rag.generators import (
    HFPretrainedModelGenerator as HFPretrainedModelGenerator,
)
from fed_rag.generators import (
    UnslothFastModelGenerator as UnslothFastModelGenerator,
)
from fed_rag.retrievers import (
    HFSentenceTransformerRetriever as HFSentenceTransformerRetriever,
)
from fed_rag.trainer_managers import (
    HuggingFaceRAGTrainerManager as HuggingFaceRAGTrainerManager,
)
from fed_rag.trainer_managers import (
    PyTorchRAGTrainerManager as PyTorchRAGTrainerManager,
)
from fed_rag.trainers import (
    HuggingFaceTrainerForLSR as HuggingFaceTrainerForLSR,
)
from fed_rag.trainers import (
    HuggingFaceTrainerForRALT as HuggingFaceTrainerForRALT,
)



"""Unsloth FastModel Generator"""

from typing import TYPE_CHECKING, Any, Optional, Union

from pydantic import ConfigDict, Field, PrivateAttr, model_validator
from typing_extensions import Self

if TYPE_CHECKING:  # pragma: no cover
    from transformers import PreTrainedModel, PreTrainedTokenizer
    from peft import PeftModel
    from transformers.generation.utils import GenerationConfig

from fed_rag.base.generator import BaseGenerator
from fed_rag.exceptions import GeneratorError
from fed_rag.tokenizers.unsloth_pretrained_tokenizer import (
    UnslothPretrainedTokenizer,
)

from .mixin import UnslothGeneratorMixin
from .utils import check_unsloth_installed

DEFAULT_PROMPT_TEMPLATE = """
You are a helpful assistant. Given the user's query, provide a succinct
and accurate response. If context is provided, use it in your answer if it helps
you to create the most accurate response.

<query>
{query}
</query>

<context>
{context}
</context>

<response>

"""


class UnslothFastModelGenerator(UnslothGeneratorMixin, BaseGenerator):
    """Unsloth FastModel Integration.

    System Requirements (see https://docs.unsloth.ai/get-started/beginner-start-here/unsloth-requirements)
        - Operating System: Works on Linux and Windows.
        - Supports NVIDIA GPUs since 2018+. Minimum CUDA Capability 7.0
        (V100, T4, Titan V, RTX 20, 30, 40x, A100, H100, L40 etc)
        Check your GPU! GTX 1070, 1080 works, but is slow.
        - Your device must have xformers, torch, BitsandBytes and triton support.
        - Unsloth only works if you have a NVIDIA GPU. Make sure you also have disk space to train & save your model
    """

    model_config = ConfigDict(protected_namespaces=("pydantic_model_",))
    model_name: str = Field(
        description="Name of Unsloth model. Used for loading the model from HF hub or local."
    )
    generation_config: "GenerationConfig" = Field(
        description="The generation config used for generating with the PreTrainedModel."
    )
    load_model_kwargs: dict = Field(
        description="Optional kwargs dict for loading ~unsloth.FastModel.from_pretrained(). Defaults to None.",
        default_factory=dict,
    )
    _prompt_template: str = PrivateAttr(default=DEFAULT_PROMPT_TEMPLATE)
    _model: Optional[Union["PreTrainedModel", "PeftModel"]] = PrivateAttr(
        default=None
    )
    _tokenizer: UnslothPretrainedTokenizer | None = PrivateAttr(default=None)

    def __init__(
        self,
        model_name: str,
        generation_config: Optional["GenerationConfig"] = None,
        prompt_template: str | None = None,
        load_model_kwargs: dict | None = None,
        load_model_at_init: bool = True,
    ):
        # if reaches here, then passed checks for extra
        from transformers.generation.utils import GenerationConfig

        generation_config = (
            generation_config if generation_config else GenerationConfig()
        )
        super().__init__(
            model_name=model_name,
            generation_config=generation_config,
            load_model_kwargs=load_model_kwargs if load_model_kwargs else {},
        )
        self._prompt_template = (
            prompt_template if prompt_template else DEFAULT_PROMPT_TEMPLATE
        )
        if load_model_at_init:
            self._model, tokenizer = self._load_model_and_tokenizer()
            self._tokenizer = UnslothPretrainedTokenizer(
                model_name=self.model_name, tokenizer=tokenizer
            )

    @model_validator(mode="before")
    @classmethod
    def check_dependencies(cls, data: Any) -> Any:
        """Validate that qdrant dependencies are installed."""
        check_unsloth_installed(cls.__name__)
        return data

    def _load_model_and_tokenizer(
        self, **kwargs: Any
    ) -> tuple[Union["PreTrainedModel", "PeftModel"], "PreTrainedTokenizer"]:
        from unsloth import FastLanguageModel

        load_kwargs = self.load_model_kwargs
        load_kwargs.update(kwargs)
        self.load_model_kwargs = load_kwargs
        model, tokenizer = FastLanguageModel.from_pretrained(
            self.model_name, **load_kwargs
        )
        return model, tokenizer

    @property
    def model(self) -> Union["PreTrainedModel", "PeftModel"]:
        if self._model is None:
            # load HF Pretrained Model
            model, tokenizer = self._load_model_and_tokenizer()
            self._model = model
            if self._tokenizer is None:
                self._tokenizer = UnslothPretrainedTokenizer(
                    model_name=self.model_name, tokenizer=tokenizer
                )
        return self._model

    @model.setter
    def model(self, value: Union["PreTrainedModel", "PeftModel"]) -> None:
        self._model = value

    @property
    def tokenizer(self) -> UnslothPretrainedTokenizer:
        return self._tokenizer

    @tokenizer.setter
    def tokenizer(self, value: UnslothPretrainedTokenizer) -> None:
        self._tokenizer = value

    @property
    def prompt_template(self) -> str:
        return self._prompt_template

    @prompt_template.setter
    def prompt_template(self, value: str) -> None:
        self._prompt_template = value

    def _get_peft_model(self, **kwargs: Any) -> "PeftModel":
        """A light wrapper over ~FastModel.get_peft_model()."""
        from unsloth import FastLanguageModel

        model = FastLanguageModel.get_peft_model(self.model, **kwargs)

        # Fix any potential dtype mismatch with any adapters and base model
        base_dtype = next(model.parameters()).dtype

        for _name, param in model.named_parameters():
            if param.requires_grad and param.dtype != base_dtype:
                param.data = param.data.to(base_dtype)

        return model

    def to_peft(self, **kwargs: Any) -> Self:
        """Sets the current model to PeftModel

        NOTE: Pass params to underlying get_peft_model using **kwargs.

        This returns Self to support fluent style:
            `generator = UnslothFastModelGenerator(...).to_peft(...)`
        """
        from peft import PeftModel

        if isinstance(self.model, PeftModel):
            raise GeneratorError(
                "Cannot use `to_peft` when underlying model is already a `~peft.PeftModel`."
            )

        # set model to new peft model
        self.model = self._get_peft_model(**kwargs)
        return self



"""Tests for Natural Questions benchmark"""

import re
import sys
from unittest.mock import MagicMock, patch

import pytest
from datasets import Dataset

import fed_rag.evals.benchmarks as benchmarks
from fed_rag.data_structures.evals import BenchmarkExample
from fed_rag.exceptions import MissingExtraError


@pytest.fixture
def dummy_natural_questions_short_answer() -> Dataset:
    """Create a dummy Natural Questions dataset with short answer."""
    return Dataset.from_dict(
        {
            "id": ["5655493461695674962"],
            "document": [
                {
                    "title": "Parma Heights, Ohio",
                    "url": "https://en.wikipedia.org/wiki/Parma_Heights,_Ohio",
                    "tokens": {
                        "token": [
                            "Parma",
                            "Heights",
                            "is",
                            "a",
                            "city",
                            "in",
                            "Cuyahoga",
                            "County",
                            ",",
                            "Ohio",
                            ".",
                        ],
                        "is_html": [
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                        ],
                        "start_byte": [
                            0,
                            6,
                            14,
                            17,
                            19,
                            24,
                            27,
                            36,
                            42,
                            44,
                            48,
                        ],
                        "end_byte": [
                            5,
                            13,
                            16,
                            18,
                            23,
                            26,
                            35,
                            42,
                            43,
                            48,
                            49,
                        ],
                    },
                }
            ],
            "question": [
                {
                    "text": "what county is parma heights ohio in",
                    "tokens": [
                        "what",
                        "county",
                        "is",
                        "parma",
                        "heights",
                        "ohio",
                        "in",
                    ],
                }
            ],
            "long_answer_candidates": [
                {
                    "start_byte": [0],
                    "end_byte": [49],
                    "start_token": [0],
                    "end_token": [10],
                    "top_level": [True],
                }
            ],
            "annotations": [
                {
                    "id": ["4569326352834458124"],
                    "long_answer": [
                        {
                            "start_byte": 0,
                            "end_byte": 49,
                            "start_token": 0,
                            "end_token": 10,
                            "candidate_index": 0,
                        }
                    ],
                    "short_answers": [
                        {
                            "start_byte": [27],
                            "end_byte": [42],
                            "start_token": [6],
                            "end_token": [7],
                            "text": ["Cuyahoga County"],
                        }
                    ],
                    "yes_no_answer": [-1],
                }
            ],
        }
    )


@pytest.fixture
def dummy_natural_questions_yes_no() -> Dataset:
    """Create a dummy Natural Questions dataset with yes/no answer."""
    return Dataset.from_dict(
        {
            "id": ["1"],
            "document": [
                {
                    "title": "Test",
                    "tokens": {
                        "token": ["Test", "document", "text", "."],
                        "is_html": [False, False, False, False],
                        "start_byte": [0, 5, 14, 18],
                        "end_byte": [4, 13, 18, 19],
                    },
                }
            ],
            "question": [
                {
                    "text": "Is this a test?",
                    "tokens": ["is", "this", "a", "test"],
                }
            ],
            "annotations": [
                {
                    "id": ["123"],
                    "long_answer": [
                        {
                            "candidate_index": -1,
                            "start_byte": -1,
                            "end_byte": -1,
                            "start_token": -1,
                            "end_token": -1,
                        }
                    ],
                    "short_answers": [
                        {
                            "start_byte": [],
                            "end_byte": [],
                            "start_token": [],
                            "end_token": [],
                            "text": [],
                        }
                    ],
                    "yes_no_answer": [1],  # YES
                }
            ],
        }
    )


@pytest.fixture
def dummy_natural_questions_no_answer() -> Dataset:
    """Create a dummy Natural Questions dataset with no answer."""
    return Dataset.from_dict(
        {
            "id": ["2"],
            "document": [
                {
                    "title": "Test",
                    "tokens": {
                        "token": ["Test", "document", "."],
                        "is_html": [False, False, False],
                        "start_byte": [0, 5, 13],
                        "end_byte": [4, 13, 14],
                    },
                }
            ],
            "question": [
                {
                    "text": "What is unknown?",
                    "tokens": ["what", "is", "unknown"],
                }
            ],
            "annotations": [
                {
                    "id": ["456"],
                    "long_answer": [
                        {
                            "candidate_index": -1,
                            "start_byte": -1,
                            "end_byte": -1,
                            "start_token": -1,
                            "end_token": -1,
                        }
                    ],
                    "short_answers": [
                        {
                            "start_byte": [],
                            "end_byte": [],
                            "start_token": [],
                            "end_token": [],
                            "text": [],
                        }
                    ],
                    "yes_no_answer": [-1],  # No yes/no answer
                }
            ],
        }
    )


@pytest.fixture
def dummy_natural_questions_multiple_short_answers() -> Dataset:
    """Create a dummy Natural Questions dataset with multiple short answers."""
    return Dataset.from_dict(
        {
            "id": ["3"],
            "document": [
                {
                    "title": "Google Founders",
                    "tokens": {
                        "token": [
                            "Google",
                            "was",
                            "founded",
                            "by",
                            "Larry",
                            "Page",
                            "and",
                            "Sergey",
                            "Brin",
                            ".",
                        ],
                        "is_html": [
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                            False,
                        ],
                        "start_byte": [0, 7, 11, 19, 22, 28, 33, 37, 44, 48],
                        "end_byte": [6, 10, 18, 21, 27, 32, 36, 43, 48, 49],
                    },
                }
            ],
            "question": [
                {
                    "text": "who founded google",
                    "tokens": ["who", "founded", "google"],
                }
            ],
            "annotations": [
                {
                    "id": ["789"],
                    "long_answer": [
                        {
                            "candidate_index": -1,
                            "start_byte": -1,
                            "end_byte": -1,
                            "start_token": -1,
                            "end_token": -1,
                        }
                    ],
                    "short_answers": [
                        {
                            "start_byte": [22, 37],
                            "end_byte": [32, 48],
                            "start_token": [4, 7],
                            "end_token": [5, 8],
                            "text": ["Larry Page", "Sergey Brin"],
                        }
                    ],
                    "yes_no_answer": [-1],
                }
            ],
        }
    )


@patch("datasets.load_dataset")
def test_natural_questions_benchmark(mock_load_dataset: MagicMock) -> None:
    # arrange & act
    nq = benchmarks.HuggingFaceNaturalQuestions()

    # assert
    mock_load_dataset.assert_called_once_with(
        benchmarks.HuggingFaceNaturalQuestions.dataset_name,
        name=nq.configuration_name,
        split=nq.split,  # Should be 'validation'
        streaming=nq.streaming,
    )


@patch("datasets.load_dataset")
def test_natural_questions_short_answer(
    mock_load_dataset: MagicMock, dummy_natural_questions_short_answer: Dataset
) -> None:
    # arrange
    mock_load_dataset.return_value = dummy_natural_questions_short_answer
    nq = benchmarks.HuggingFaceNaturalQuestions()

    # assert
    assert isinstance(nq[0], BenchmarkExample)
    assert nq[0].query == "what county is parma heights ohio in"
    assert nq[0].response == "Cuyahoga County"
    assert (
        nq[0].context == "Parma Heights is a city in Cuyahoga County , Ohio ."
    )


@patch("datasets.load_dataset")
def test_natural_questions_multiple_short_answers(
    mock_load_dataset: MagicMock,
    dummy_natural_questions_multiple_short_answers: Dataset,
) -> None:
    # arrange
    mock_load_dataset.return_value = (
        dummy_natural_questions_multiple_short_answers
    )
    nq = benchmarks.HuggingFaceNaturalQuestions()

    # assert
    assert nq[0].response == "Larry Page and Sergey Brin"


@patch("datasets.load_dataset")
def test_natural_questions_yes_no_answer(
    mock_load_dataset: MagicMock, dummy_natural_questions_yes_no: Dataset
) -> None:
    # arrange
    mock_load_dataset.return_value = dummy_natural_questions_yes_no
    nq = benchmarks.HuggingFaceNaturalQuestions()

    # assert
    assert nq[0].response == "YES"


@patch("datasets.load_dataset")
def test_natural_questions_no_answer(
    mock_load_dataset: MagicMock, dummy_natural_questions_no_answer: Dataset
) -> None:
    # arrange
    mock_load_dataset.return_value = dummy_natural_questions_no_answer
    nq = benchmarks.HuggingFaceNaturalQuestions()

    # assert
    assert nq[0].response == "[NO ANSWER]"


@patch("datasets.load_dataset")
def test_natural_questions_long_answer_only(
    mock_load_dataset: MagicMock,
) -> None:
    """Test handling when only long answer exists."""
    dataset = Dataset.from_dict(
        {
            "id": ["4"],
            "document": [
                {
                    "title": "Test",
                    "tokens": {
                        "token": ["Long", "document", "text", "."],
                        "is_html": [False, False, False, False],
                        "start_byte": [0, 5, 14, 18],
                        "end_byte": [4, 13, 18, 19],
                    },
                }
            ],
            "question": [
                {
                    "text": "What is this about?",
                    "tokens": ["what", "is", "this", "about"],
                }
            ],
            "annotations": [
                {
                    "id": ["789"],
                    "long_answer": [
                        {
                            "candidate_index": 0,
                            "start_byte": 0,
                            "end_byte": 19,
                            "start_token": 0,
                            "end_token": 3,
                        }
                    ],  # Valid long answer
                    "short_answers": [
                        {
                            "start_byte": [],
                            "end_byte": [],
                            "start_token": [],
                            "end_token": [],
                            "text": [],
                        }
                    ],
                    "yes_no_answer": [-1],
                }
            ],
        }
    )
    mock_load_dataset.return_value = dataset
    nq = benchmarks.HuggingFaceNaturalQuestions()
    assert nq[0].response == "[LONG ANSWER EXISTS]"


@patch("datasets.load_dataset")
def test_natural_questions_no_annotations(
    mock_load_dataset: MagicMock,
) -> None:
    """Test handling when annotations are missing."""
    dataset = Dataset.from_dict(
        {
            "id": ["5"],
            "document": [{"title": "Test"}],  # No tokens
            "question": [{"text": "Test question?"}],
            "annotations": [{}],  # Empty annotations
        }
    )
    mock_load_dataset.return_value = dataset
    nq = benchmarks.HuggingFaceNaturalQuestions()
    assert nq[0].response == "[NO ANSWER]"
    assert nq[0].context == "Test"  # Falls back to title


@patch("datasets.load_dataset")
def test_natural_questions_missing_fields(
    mock_load_dataset: MagicMock,
) -> None:
    """Test handling when various fields are missing."""
    dataset = Dataset.from_dict(
        {
            "id": ["6"],
            "document": [{}],  # Empty document
            "question": [{}],  # Empty question
            "annotations": [{"yes_no_answer": [-1]}],
        }
    )
    mock_load_dataset.return_value = dataset
    nq = benchmarks.HuggingFaceNaturalQuestions()
    assert nq[0].query == ""
    assert nq[0].response == "[NO ANSWER]"
    assert nq[0].context is None


def test_huggingface_evals_extra_missing() -> None:
    """Test that proper error is raised when huggingface-evals extra is missing."""
    modules = {
        "datasets": None,
    }
    module_to_import = "fed_rag.evals.benchmarks"
    original_module = sys.modules.pop(module_to_import, None)

    with patch.dict("sys.modules", modules):
        msg = (
            "`HuggingFaceNaturalQuestions` requires the `huggingface-evals` extra to be installed. "
            "To fix please run `pip install fed-rag[huggingface-evals]`."
        )
        with pytest.raises(
            MissingExtraError,
            match=re.escape(msg),
        ):
            import fed_rag.evals.benchmarks as benchmarks

            benchmarks.HuggingFaceNaturalQuestions()

    # restore module so to not affect other tests
    if original_module:
        sys.modules[module_to_import] = original_module



from fed_rag import NoEncodeRAGSystem, RAGSystem
from fed_rag.exceptions import FedRAGError


def _validate_rag_system(rag_system: RAGSystem | NoEncodeRAGSystem) -> None:
    # Skip validation if environment variable is set
    import os

    if os.environ.get("FEDRAG_SKIP_VALIDATION") == "1":
        return

    from fed_rag.generators.huggingface import (
        HFPeftModelGenerator,
        HFPretrainedModelGenerator,
    )
    from fed_rag.generators.unsloth import UnslothFastModelGenerator
    from fed_rag.retrievers.huggingface.hf_sentence_transformer import (
        HFSentenceTransformerRetriever,
    )

    if not isinstance(
        rag_system.generator,
        (
            HFPretrainedModelGenerator,
            HFPeftModelGenerator,
            UnslothFastModelGenerator,
        ),
    ):
        raise FedRAGError(
            "Generator must be HFPretrainedModelGenerator or HFPeftModelGenerator"
        )

    if isinstance(rag_system, RAGSystem) and not isinstance(
        rag_system.retriever, HFSentenceTransformerRetriever
    ):
        raise FedRAGError("Retriever must be a HFSentenceTransformerRetriever")



from unittest.mock import AsyncMock, patch

import pytest
from mcp import StdioServerParameters
from mcp.types import CallToolResult, ImageContent, TextContent

from fed_rag.base.no_encode_knowledge_store import (
    BaseAsyncNoEncodeKnowledgeStore,
)
from fed_rag.data_structures import KnowledgeNode
from fed_rag.exceptions import MCPKnowledgeStoreError
from fed_rag.knowledge_stores.no_encode import (
    MCPKnowledgeStore,
    MCPStdioKnowledgeSource,
    MCPStreamableHttpKnowledgeSource,
)


@pytest.fixture
def mcp_streamable_http_source() -> MCPStreamableHttpKnowledgeSource:
    return MCPStreamableHttpKnowledgeSource(
        url="https://fake-mcp-url.io",
        tool_name="fake_tool",
        query_param_name="query",
        tool_call_kwargs={"param_1": 1, "param_2": "value 2"},
    )


@pytest.fixture
def mcp_stdio_source() -> MCPStdioKnowledgeSource:
    return MCPStdioKnowledgeSource(
        server_params=StdioServerParameters(
            command="uv run",
            args=["fake.py"],
        ),
        tool_name="fake_tool",
        query_param_name="query",
        tool_call_kwargs={"param_1": 1, "param_2": "value 2"},
    )


@pytest.fixture
def call_tool_result() -> CallToolResult:
    content = [
        TextContent(text="text 1", type="text"),
        TextContent(text="text 2", type="text"),
        ImageContent(data="fakeimage", mimeType="image/png", type="image"),
    ]
    return CallToolResult(content=content)


def test_mcp_knowledge_store_class() -> None:
    names_of_base_classes = [b.__name__ for b in MCPKnowledgeStore.__mro__]
    assert BaseAsyncNoEncodeKnowledgeStore.__name__ in names_of_base_classes


def test_mcp_knowledge_store_init(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    assert store.name == "default-mcp"
    assert store.sources == {
        mcp_streamable_http_source.name: mcp_streamable_http_source
    }


@pytest.mark.asyncio
@patch(
    "fed_rag.knowledge_stores.no_encode.mcp.sources.streamable_http.ClientSession"
)
@patch(
    "fed_rag.knowledge_stores.no_encode.mcp.sources.streamable_http.streamablehttp_client"
)
async def test_mcp_knowledge_store_retrieve_streamable_http(
    mock_streamable_client: AsyncMock,
    mock_session_class: AsyncMock,
    mcp_streamable_http_source: MCPStdioKnowledgeSource,
    call_tool_result: CallToolResult,
) -> None:
    # arrange mocks
    mock_session_instance = AsyncMock()
    mock_session_instance.initialize = AsyncMock()
    mock_session_instance.call_tool = AsyncMock(return_value=call_tool_result)

    # mock context managers
    mock_session_class.return_value.__aenter__ = AsyncMock(
        return_value=mock_session_instance
    )
    mock_session_class.return_value.__aexit__ = AsyncMock(return_value=None)

    mock_streamable_client.return_value.__aenter__ = AsyncMock(
        return_value=(None, None, None)
    )
    mock_streamable_client.return_value.__aexit__ = AsyncMock(
        return_value=None
    )

    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)
    result = await store.retrieve(query="mock_query", top_k=10)

    assert result[0][0] == 1.0  # Default similiarity score
    assert result[1][0] == 1.0  # Default similiarity score
    assert result[0][1].text_content == "text 1"
    assert result[1][1].text_content == "text 2"
    assert result[0][1].metadata == mcp_streamable_http_source.model_dump()
    assert result[1][1].metadata == mcp_streamable_http_source.model_dump()
    mock_session_instance.call_tool.assert_called_once_with(
        mcp_streamable_http_source.tool_name,
        arguments={
            mcp_streamable_http_source.query_param_name: "mock_query",
            **mcp_streamable_http_source.tool_call_kwargs,
        },
    )


@pytest.mark.asyncio
@patch("fed_rag.knowledge_stores.no_encode.mcp.sources.stdio.ClientSession")
@patch("fed_rag.knowledge_stores.no_encode.mcp.sources.stdio.stdio_client")
async def test_mcp_knowledge_store_retrieve_stdio(
    mock_stdio_client: AsyncMock,
    mock_session_class: AsyncMock,
    mcp_stdio_source: MCPStreamableHttpKnowledgeSource,
    call_tool_result: CallToolResult,
) -> None:
    # arrange mocks
    mock_session_instance = AsyncMock()
    mock_session_instance.initialize = AsyncMock()
    mock_session_instance.call_tool = AsyncMock(return_value=call_tool_result)

    # mock context managers
    mock_session_class.return_value.__aenter__ = AsyncMock(
        return_value=mock_session_instance
    )
    mock_session_class.return_value.__aexit__ = AsyncMock(return_value=None)

    mock_stdio_client.return_value.__aenter__ = AsyncMock(
        return_value=(None, None)
    )
    mock_stdio_client.return_value.__aexit__ = AsyncMock(return_value=None)

    store = MCPKnowledgeStore().add_source(mcp_stdio_source)
    result = await store.retrieve(query="mock_query", top_k=10)

    assert result[0][0] == 1.0  # Default similiarity score
    assert result[1][0] == 1.0  # Default similiarity score
    assert result[0][1].text_content == "text 1"
    assert result[1][1].text_content == "text 2"
    assert result[0][1].metadata == mcp_stdio_source.model_dump()
    assert result[1][1].metadata == mcp_stdio_source.model_dump()
    mock_session_instance.call_tool.assert_called_once_with(
        mcp_stdio_source.tool_name,
        arguments={
            mcp_stdio_source.query_param_name: "mock_query",
            **mcp_stdio_source.tool_call_kwargs,
        },
    )


@pytest.mark.asyncio
@patch("fed_rag.knowledge_stores.no_encode.mcp.sources.stdio.ClientSession")
@patch("fed_rag.knowledge_stores.no_encode.mcp.sources.stdio.stdio_client")
async def test_mcp_knowledge_store_retrieve_with_reranker(
    mock_stdio_client: AsyncMock,
    mock_session_class: AsyncMock,
    mcp_stdio_source: MCPStreamableHttpKnowledgeSource,
    call_tool_result: CallToolResult,
) -> None:
    # arrange mocks
    mock_session_instance = AsyncMock()
    mock_session_instance.initialize = AsyncMock()
    mock_session_instance.call_tool = AsyncMock(return_value=call_tool_result)

    # mock context managers
    mock_session_class.return_value.__aenter__ = AsyncMock(
        return_value=mock_session_instance
    )
    mock_session_class.return_value.__aexit__ = AsyncMock(return_value=None)

    mock_stdio_client.return_value.__aenter__ = AsyncMock(
        return_value=(None, None)
    )
    mock_stdio_client.return_value.__aexit__ = AsyncMock(return_value=None)

    store = (
        MCPKnowledgeStore()
        .add_source(mcp_stdio_source)
        .with_reranker(
            lambda nodes, query: [
                (len(query) / len(n.text_content), n) for n in nodes
            ]
        )
    )

    result = await store.retrieve(query="mock_query", top_k=10)

    assert result[0][0] == len("mock_query") / len(result[0][1].text_content)
    assert result[1][0] == len("mock_query") / len(result[1][1].text_content)
    assert result[0][1].text_content == "text 1"
    assert result[1][1].text_content == "text 2"
    assert result[0][1].metadata == mcp_stdio_source.model_dump()
    assert result[1][1].metadata == mcp_stdio_source.model_dump()
    mock_session_instance.call_tool.assert_called_once_with(
        mcp_stdio_source.tool_name,
        arguments={
            mcp_stdio_source.query_param_name: "mock_query",
            **mcp_stdio_source.tool_call_kwargs,
        },
    )


@pytest.mark.asyncio
async def test_add_unsupported_source_type_raises_error() -> None:
    with pytest.raises(
        MCPKnowledgeStoreError,
        match="Cannot add source of type: <class 'int'>",
    ):
        _ = MCPKnowledgeStore().add_source(1)


def test_add_source_raises_error_with_existing_name(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    with pytest.raises(
        MCPKnowledgeStoreError,
        match=f"A source with the same name, {mcp_streamable_http_source.name}, already exists.",
    ):
        _ = (
            MCPKnowledgeStore()
            .add_source(mcp_streamable_http_source)
            .add_source(mcp_streamable_http_source)
        )


# test not implemented
@pytest.mark.asyncio
async def test_load_node_raises_not_implemented_error(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    with pytest.raises(NotImplementedError):
        await store.load_node(
            KnowledgeNode(node_type="text", text_content="text 1")
        )


@pytest.mark.asyncio
async def test_load_nodes_raises_not_implemented_error(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    with pytest.raises(NotImplementedError):
        await store.load_nodes(
            [KnowledgeNode(node_type="text", text_content="text 1")]
        )


@pytest.mark.asyncio
async def test_delete_node_raises_not_implemented_error(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    with pytest.raises(NotImplementedError):
        await store.delete_node("")


@pytest.mark.asyncio
async def test_clear_raises_not_implemented_error(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    with pytest.raises(NotImplementedError):
        await store.clear()


def test_count_raises_not_implemented_error(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    with pytest.raises(NotImplementedError):
        store.count


def test_persist_raises_not_implemented_error(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    with pytest.raises(NotImplementedError):
        store.persist()


def test_load_raises_not_implemented_error(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    with pytest.raises(NotImplementedError):
        store.load()


@pytest.mark.asyncio
async def test_batch_retrieve_raises_not_implemented_error(
    mcp_streamable_http_source: MCPStreamableHttpKnowledgeSource,
) -> None:
    store = MCPKnowledgeStore().add_source(mcp_streamable_http_source)

    with pytest.raises(NotImplementedError):
        await store.batch_retrieve(queries=["query 1", "query 2"])



"""Exceptions for Evals."""

from .core import FedRAGError, FedRAGWarning


class EvalsError(FedRAGError):
    """Base evals error for all evals-related exceptions."""

    pass


class EvalsWarning(FedRAGWarning):
    """Base inspector warning for all evals-related warnings."""

    pass


class BenchmarkGetExamplesError(EvalsError):
    """Raised if an error occurs when getting examples for a benchmark."""

    pass


class BenchmarkParseError(EvalsError):
    """Raised when errors occur during parsing examples."""

    pass


class EvaluationsFileNotFoundError(EvalsError, FileNotFoundError):
    """Benchmark evaluations file not found error."""

    pass



"""Qdrant utils module."""

from importlib.util import find_spec
from typing import TYPE_CHECKING

from fed_rag.data_structures.knowledge_node import KnowledgeNode
from fed_rag.exceptions import KnowledgeStoreError, MissingExtraError

if TYPE_CHECKING:  # pragma: no cover
    from qdrant_client.http.models import ScoredPoint
    from qdrant_client.models import PointStruct


def check_qdrant_installed() -> None:
    if find_spec("qdrant_client") is None:
        raise MissingExtraError(
            "Qdrant knowledge stores require the qdrant-client to be installed. "
            "To fix please run `pip install fed-rag[qdrant]`."
        )


def convert_knowledge_node_to_qdrant_point(
    node: KnowledgeNode,
) -> "PointStruct":
    from qdrant_client.models import PointStruct

    if node.embedding is None:
        raise KnowledgeStoreError(
            "Cannot load a node with embedding set to None."
        )

    return PointStruct(
        id=node.node_id,
        vector=node.embedding,
        payload=node.model_dump_without_embeddings(),
    )


def convert_scored_point_to_knowledge_node_and_score_tuple(
    scored_point: "ScoredPoint",
) -> tuple[float, KnowledgeNode]:
    knowledge_data = scored_point.payload
    knowledge_data.update(
        embedding=scored_point.vector
    )  # attach vector to embedding if it is even returned
    return (
        scored_point.score,
        KnowledgeNode.model_validate(knowledge_data),
    )



import tempfile
from pathlib import Path

import pytest
import torch

from fed_rag.base.knowledge_store import BaseKnowledgeStore
from fed_rag.data_structures.knowledge_node import KnowledgeNode
from fed_rag.exceptions import KnowledgeStoreNotFoundError
from fed_rag.knowledge_stores.in_memory import InMemoryKnowledgeStore


@pytest.fixture
def text_nodes() -> list[KnowledgeNode]:
    return [
        KnowledgeNode(
            embedding=[1.0, 0.0, 1.0],
            node_type="text",
            text_content="node 1",
            metadata={"key1": "value1"},
        ),
        KnowledgeNode(
            embedding=[1.0, 0.0, 0.0],
            node_type="text",
            text_content="node 2",
            metadata={"key2": "value2"},
        ),
        KnowledgeNode(
            embedding=[1.0, 1.0, 0.0],
            node_type="text",
            text_content="node 3",
            metadata={"key3": "value3"},
        ),
    ]


def test_in_memory_knowledge_store_class() -> None:
    names_of_base_classes = [
        b.__name__ for b in InMemoryKnowledgeStore.__mro__
    ]
    assert BaseKnowledgeStore.__name__ in names_of_base_classes


def test_in_memory_knowledge_store_init() -> None:
    knowledge_store = InMemoryKnowledgeStore()

    assert knowledge_store.count == 0


def test_from_nodes(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=text_nodes)

    assert knowledge_store.count == 3
    assert all(n.node_id in knowledge_store._data for n in text_nodes)


def test_delete_node(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=text_nodes)

    assert knowledge_store.count == 3

    res = knowledge_store.delete_node(text_nodes[0].node_id)

    assert res is True
    assert knowledge_store.count == 2
    assert text_nodes[0].node_id not in knowledge_store._data


def test_delete_node_returns_false(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=text_nodes)

    assert knowledge_store.count == 3

    res = knowledge_store.delete_node("non_included_id")

    assert res is False
    assert knowledge_store.count == 3
    assert all(n.node_id in knowledge_store._data for n in text_nodes)


def test_load_node(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = InMemoryKnowledgeStore()
    assert knowledge_store.count == 0

    knowledge_store.load_node(text_nodes[-1])

    assert knowledge_store.count == 1
    assert text_nodes[-1].node_id in knowledge_store._data


def test_load_nodes(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = InMemoryKnowledgeStore()
    assert knowledge_store.count == 0

    knowledge_store.load_nodes(text_nodes)

    assert knowledge_store.count == 3
    assert all(n.node_id in knowledge_store._data for n in text_nodes)


@pytest.mark.parametrize(
    ("query_emb", "top_k", "expected_node_ix"),
    [([1.0, 1.0, 1.0], 2, [0, 2]), ([0.5, 0.0, 0.0], 1, [1])],
    ids=[str([1.0, 1.0, 1.0]), str([0.5, 0.0, 0.0])],
)
def test_load_and_offload_nodes(
    query_emb: list[float],
    top_k: int,
    expected_node_ix: list[int],
    text_nodes: list[KnowledgeNode],
) -> None:
    knowledge_store = InMemoryKnowledgeStore()
    assert knowledge_store.count == 0

    knowledge_store.load_nodes(text_nodes)
    # act
    res = knowledge_store.retrieve(query_emb, top_k=top_k)
    # assert
    assert [el[1] for el in res] == [text_nodes[ix] for ix in expected_node_ix]
    # print(isinstance(knowledge_store._data_storage, torch.Tensor))
    knowledge_store.delete_node(text_nodes[0].node_id)
    assert knowledge_store.count == 2
    res = knowledge_store.retrieve(query_emb, top_k=top_k)
    node = KnowledgeNode(
        embedding=[1.0, 1.0, 0.0],
        node_type="text",
        text_content="node 4",
        metadata={"key4": "value4"},
    )
    knowledge_store.load_nodes([node])
    assert not isinstance(knowledge_store._data_storage, torch.Tensor)


def test_clear(text_nodes: list[KnowledgeNode]) -> None:
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=text_nodes)
    assert knowledge_store.count == 3

    knowledge_store.clear()

    assert knowledge_store.count == 0
    assert all(n.node_id not in knowledge_store._data for n in text_nodes)


@pytest.mark.parametrize(
    ("query_emb", "top_k", "expected_node_ix"),
    [([1.0, 1.0, 1.0], 2, [0, 2]), ([0.5, 0.0, 0.0], 1, [1])],
    ids=[str([1.0, 1.0, 1.0]), str([0.5, 0.0, 0.0])],
)
def test_retrieve(
    query_emb: list[float],
    top_k: int,
    expected_node_ix: list[int],
    text_nodes: list[KnowledgeNode],
) -> None:
    # arrange
    knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=text_nodes)

    # act
    res = knowledge_store.retrieve(query_emb, top_k=top_k)

    # assert
    assert [el[1] for el in res] == [text_nodes[ix] for ix in expected_node_ix]


def test_persist(text_nodes: list[KnowledgeNode]) -> None:
    with tempfile.TemporaryDirectory() as dirpath:
        knowledge_store = InMemoryKnowledgeStore.from_nodes(nodes=text_nodes)
        knowledge_store.cache_dir = dirpath
        knowledge_store.persist()

        filename = Path(dirpath) / f"{knowledge_store.name}.parquet"
        assert filename.exists()


def test_load(text_nodes: list[KnowledgeNode]) -> None:
    with tempfile.TemporaryDirectory() as dirpath:
        knowledge_store = InMemoryKnowledgeStore.from_nodes(
            nodes=text_nodes, name="test_ks", cache_dir=dirpath
        )
        knowledge_store.persist()

        # load into new empty instance
        loaded_knowledge_store = InMemoryKnowledgeStore(
            name="test_ks", cache_dir=dirpath
        )
        loaded_knowledge_store.load()

        assert loaded_knowledge_store._data == knowledge_store._data


def test_load_with_missing_file_raises_error() -> None:
    with tempfile.TemporaryDirectory() as dirpath:
        knowledge_store = InMemoryKnowledgeStore(cache_dir=dirpath)

        with pytest.raises(KnowledgeStoreNotFoundError):
            knowledge_store.load()


def test_persist_overwrite(text_nodes: list[KnowledgeNode]) -> None:
    with tempfile.TemporaryDirectory() as dirpath:
        knowledge_store = InMemoryKnowledgeStore.from_nodes(
            nodes=text_nodes, name="test_ks", cache_dir=dirpath
        )
        knowledge_store.persist()

        # overwrite
        knowledge_store.load_node(
            KnowledgeNode(
                embedding=[1.0, 1.0, 1.0],
                node_type="text",
                text_content="node 4",
                metadata={"key4": "value4"},
            )
        )
        knowledge_store.persist()

        # load into new empty instance
        loaded_knowledge_store = InMemoryKnowledgeStore(
            name="test_ks", cache_dir=dirpath
        )
        loaded_knowledge_store.load()

        assert loaded_knowledge_store._data == knowledge_store._data


def test_batch_retrieve_raises_error() -> None:
    with tempfile.TemporaryDirectory() as dirpath:
        knowledge_store = InMemoryKnowledgeStore(cache_dir=dirpath)

        with pytest.raises(NotImplementedError):
            knowledge_store.batch_retrieve(query_embs=[[1, 2, 3], [4, 5, 6]])



"""SQuAD 2.0 benchmark"""

from typing import Any

from pydantic import model_validator

from fed_rag.base.evals.benchmark import BaseBenchmark

from .mixin import HuggingFaceBenchmarkMixin
from .utils import check_huggingface_evals_installed


class HuggingFaceSQuADv2(HuggingFaceBenchmarkMixin, BaseBenchmark):
    """HuggingFace SQuAD 2.0 Benchmark.

    Stanford Question Answering Dataset (SQuAD) 2.0 combines 100,000 questions
    from SQuAD 1.1 with over 50,000 unanswerable questions. Systems must not
    only answer questions when possible, but also determine when no answer is
    supported by the paragraph.

    Example schema:
        {
            "id": "56ddde2d66d3e219004dad4d",
            "title": "Symbiosis",
            "context": "Symbiotic relationships include those associations in which one organism lives on another (ectosymbiosis, such as...",
            "question": "What is an example of ectosymbiosis?",
            "answers": {
                "text": ["mistletoe"],
                "answer_start": [114]
            }
        }

    For unanswerable questions, the answers field has empty lists:
        {
            "answers": {
                "text": [],
                "answer_start": []
            }
        }
    """

    dataset_name = "squad_v2"
    configuration_name: str | None = None

    def _get_query_from_example(self, example: dict[str, Any]) -> str:
        return str(example["question"])

    def _get_response_from_example(self, example: dict[str, Any]) -> str:
        answers = example.get("answers", {})
        answer_texts = answers.get("text", [])

        if answer_texts:
            # Return the first answer (they are typically variations of the same answer)
            return str(answer_texts[0])
        else:
            # For unanswerable questions, return a special token
            return "[NO ANSWER]"

    def _get_context_from_example(self, example: dict[str, Any]) -> str:
        return str(example["context"])

    @model_validator(mode="before")
    @classmethod
    def _validate_extra_installed(cls, data: Any) -> Any:
        """Validate that huggingface-evals dependencies are installed."""
        check_huggingface_evals_installed(cls.__name__)
        return data



"""Exact Match Metric"""

from typing import Any

from fed_rag.base.evals.metric import BaseEvaluationMetric


class ExactMatchEvaluationMetric(BaseEvaluationMetric):
    """Exact match evaluation metric class."""

    def __call__(
        self, prediction: str, actual: str, *args: Any, **kwargs: Any
    ) -> float:
        return float(prediction.lower() == actual.lower())



from .boolq import HuggingFaceBoolQ
from .hellaswag import HuggingFaceHellaSwag
from .hotpotqa import HuggingFaceHotpotQA
from .mixin import HuggingFaceBenchmarkMixin
from .mmlu import HuggingFaceMMLU
from .natural_questions import HuggingFaceNaturalQuestions
from .pubmedqa import HuggingFacePubMedQA
from .squad_v2 import HuggingFaceSQuADv2

__all__ = [
    "HuggingFaceBenchmarkMixin",
    "HuggingFaceMMLU",
    "HuggingFacePubMedQA",
    "HuggingFaceHotpotQA",
    "HuggingFaceSQuADv2",
    "HuggingFaceNaturalQuestions",
    "HuggingFaceBoolQ",
    "HuggingFaceHellaSwag",
]



import re

import pytest

from fed_rag.exceptions import BenchmarkGetExamplesError

from . import _benchmarks as benchmarks


def test_sequence_interface() -> None:
    # typical pattern
    test_benchmark = benchmarks.TestBenchmark()

    assert len(test_benchmark) == 3
    assert test_benchmark.num_examples == 3
    for ix in range(len(test_benchmark)):
        assert test_benchmark[ix] == test_benchmark._examples[ix]
    example_iter = iter(test_benchmark.as_iterator())
    assert next(example_iter) == test_benchmark[0]


def test_get_example_raises_exception() -> None:
    # typical pattern

    with pytest.raises(
        BenchmarkGetExamplesError,
        match=re.escape("Failed to get examples: Too bad, so sad."),
    ):
        _ = benchmarks.TestBenchmarkBadExamples()



"""Public Core API"""

from .no_encode_rag_system import AsyncNoEncodeRAGSystem, NoEncodeRAGSystem
from .rag_system import AsyncRAGSystem, RAGSystem

__all__ = [
    "AsyncNoEncodeRAGSystem",
    "AsyncRAGSystem",
    "NoEncodeRAGSystem",
    "RAGSystem",
]



"""HuggingFaceBenchmarkMixin"""

from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    ClassVar,
    Generator,
    Optional,
    Sequence,
    Union,
    cast,
)

from pydantic import BaseModel, PrivateAttr

from fed_rag.data_structures.evals import BenchmarkExample
from fed_rag.exceptions import EvalsError

if TYPE_CHECKING:  # pragma: no cover
    from datasets import Dataset, IterableDataset


BENCHMARK_EXAMPLE_JSON_KEY = "__benchmark_example_json"


class HuggingFaceBenchmarkMixin(BaseModel, ABC):
    """Mixin for HuggingFace Benchmarks"""

    dataset_name: ClassVar[str]
    configuration_name: str | None = None
    split: str = "test"
    streaming: bool = False
    load_kwargs: dict[str, Any] = {}
    _dataset: Optional["Dataset"] = PrivateAttr(default=None)

    @abstractmethod
    def _get_query_from_example(self, example: dict[str, Any]) -> str:
        """Derive the query from the example."""

    @abstractmethod
    def _get_response_from_example(self, example: dict[str, Any]) -> str:
        """Derive the response from the example."""

    @abstractmethod
    def _get_context_from_example(self, example: dict[str, Any]) -> str | None:
        """Derive the context from the example."""

    def _map_dataset_example(self, example: dict[str, Any]) -> dict[str, Any]:
        """Map the examples in the dataset to include a `~fed_rag.data_structures.evals.BenchmarkExample`."""
        query = self._get_query_from_example(example)
        response = self._get_response_from_example(example)
        context = self._get_context_from_example(example)

        example[BENCHMARK_EXAMPLE_JSON_KEY] = {
            "query": query,
            "response": response,
            "context": context,
        }
        return example

    def _load_dataset(self) -> Union["Dataset", "IterableDataset"]:
        from datasets import load_dataset

        try:
            loaded_dataset = load_dataset(
                self.dataset_name,
                name=self.configuration_name,
                split=self.split,
                streaming=self.streaming,
                **self.load_kwargs,
            )
        except Exception as e:
            raise EvalsError(
                f"Failed to load dataset, `{self.dataset_name}`, due to error: {str(e)}"
            ) from e

        # add BenchmarkExample to dataset
        return loaded_dataset.map(self._map_dataset_example)

    @property
    def dataset(self) -> Union["Dataset", "IterableDataset"]:
        if self._dataset is None:
            self._dataset = self._load_dataset()

        return self._dataset

    # Provide required implementations for abstractmethods in ~BaseBenchmark
    def _get_examples(self, **kwargs: Any) -> Sequence[BenchmarkExample]:
        from datasets import Dataset

        if isinstance(self.dataset, Dataset):
            return [
                BenchmarkExample.model_validate(el)
                for el in self.dataset[BENCHMARK_EXAMPLE_JSON_KEY]
            ]
        else:
            return []  # examples should be streamed

    def as_stream(self) -> Generator[BenchmarkExample, None, None]:
        from datasets import IterableDataset

        # check if dataset is an iterable one
        if isinstance(self.dataset, IterableDataset):
            iterable_dataset = self.dataset
        else:
            iterable_dataset = self.dataset.to_iterable_dataset()

        # map the iterable_dataset to get the examples i
        iterable_dataset = iterable_dataset.map(self._map_dataset_example)

        for hf_example in iterable_dataset:
            yield BenchmarkExample.model_validate(
                hf_example[BENCHMARK_EXAMPLE_JSON_KEY]
            )

    @property
    def num_examples(self) -> int:
        from datasets import SplitInfo

        if splits := self.dataset.info.splits:
            try:
                split_info = splits[self.split]
                split_info = cast(
                    SplitInfo,
                    split_info,
                )
                return int(split_info.num_examples)
            except (KeyError, TypeError):
                raise EvalsError(
                    f"Unable to get size of dataset: `{self.dataset_name}`. "
                    f"Split, `{self.split}` does not exist in the splits of the dataset."
                )
        else:
            raise EvalsError(
                f"Unable to get size of dataset: `{self.dataset_name}`. "
                "The dataset does not have any listed splits."
            )



import importlib

import pytest

from fed_rag.evals import __all__ as _evals_all
from fed_rag.evals.benchmarks import __all__ as _benchmarks_all


@pytest.mark.parametrize("name", _evals_all)
def test_evals_all_importable(name: str) -> None:
    """Tests that all names listed in evals __all__ are importable."""
    mod = importlib.import_module("fed_rag.evals")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None


@pytest.mark.parametrize("name", _benchmarks_all)
def test_evals_benchmarks_all_importable(name: str) -> None:
    """Tests that all names listed in evals.benchmarks __all__ are importable."""
    mod = importlib.import_module("fed_rag.evals.benchmarks")
    attr = getattr(mod, name)

    assert hasattr(mod, name)
    assert attr is not None



"""Base Benchmark and Benchmarker"""

import contextlib
from datetime import datetime
from pathlib import Path
from typing import Any, Generator

from pydantic import BaseModel
from typing_extensions import assert_never

from fed_rag import RAGSystem
from fed_rag.base.evals.benchmark import BaseBenchmark
from fed_rag.base.evals.metric import BaseEvaluationMetric
from fed_rag.data_structures.evals import (
    AggregationMode,
    BenchmarkEvaluatedExample,
    BenchmarkExample,
    BenchmarkResult,
)

DEFAULT_OUTPUT_DIR = Path(".fed_rag") / "benchmark_results"


class Benchmarker(BaseModel):
    """Benchmarker"""

    rag_system: RAGSystem

    def _update_running_score(
        self,
        agg: AggregationMode,
        running_score: float | None,
        next_score: float,
        num_examples_seen: int,
    ) -> float:
        """Update the running score.

        Args:
            agg (AggregationMode): aggregation mode.
            running_score (float): the running score to be updated.
            next_score (float): the score of the latest scored example.
            num_examples_seen (int): the number of examples seen prior to the
                latest scored example.

        Returns:
            float: the updated running score
        """
        if not running_score:
            return next_score

        match agg:
            case AggregationMode.AVG:
                return (num_examples_seen * running_score + next_score) / (
                    num_examples_seen + 1
                )
            case AggregationMode.SUM:
                return running_score + next_score
            case AggregationMode.MAX:
                if running_score < next_score:
                    return next_score
                else:
                    return running_score
            case AggregationMode.MIN:
                if running_score > next_score:
                    return next_score
                else:
                    return running_score
            case _:  # pragma: no cover
                assert_never(agg)

    @contextlib.contextmanager
    def _get_examples_iterator(
        self, benchmark: BaseBenchmark, is_streaming: bool
    ) -> Generator[BenchmarkExample, None, None]:
        """Wrapper over the iterator or stream.

        To handle generator clean up safely.
        """
        if is_streaming:
            examples_iterator = benchmark.as_stream()
        else:
            examples_iterator = benchmark.as_iterator()

        try:
            yield examples_iterator
        finally:
            if hasattr(examples_iterator, "close"):
                examples_iterator.close()

    def run(
        self,
        benchmark: BaseBenchmark,
        metric: BaseEvaluationMetric,
        is_streaming: bool = False,
        agg: AggregationMode | str = "avg",
        batch_size: int = 1,
        num_examples: int | None = None,
        num_workers: int = 1,
        output_dir: Path | str = DEFAULT_OUTPUT_DIR,
        save_evaluations: bool = False,
        **kwargs: Any,
    ) -> BenchmarkResult:
        """Execute the benchmark using the associated `RAGSystem`.

        Args:
            agg (AggregationMode | str): the aggregation mode to apply to all example scores.
                Modes include `avg`, `sum`, `max`, or `min`.
            benchmark (BaseBenchmark): the benchmark to run the `RAGSystem` against.
            batch_size (int, optional): number of examples to process in a single batch.
            metric (BaseEvaluationMetric): the metric to use for evaluation.
            num_examples (int | None, optional): Number of examples to use from
                the benchmark. If None, then the entire collection of examples of
                the benchmark are ran. Defaults to None.
            num_workers (int, optional): concurrent execution via threads.
            output_dir (Path | None): the output directory for saving evaluations. Defaults to None.

        Returns:
            BenchmarkResult: the benchmark result

        TODO: implement concurrent as well as batch execution. Need RAGSystem
        to be able to handle batches as well.
        """
        if isinstance(output_dir, str):
            output_dir = Path(output_dir)

        # create file for saving evaluations
        if save_evaluations:
            output_dir.mkdir(parents=True, exist_ok=True)
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            filename = (
                output_dir
                / f"{benchmark.__class__.__name__}-{timestamp}.jsonl"
            )
            f = open(filename, "w")
        else:
            filename = None
            f = None

        try:
            with self._get_examples_iterator(
                benchmark, is_streaming
            ) as examples_iterator:
                running_score = None
                num_seen = 0
                for example in examples_iterator:
                    if num_seen == num_examples:
                        break

                    # prediction
                    result = self.rag_system.query(example.query)

                    # evaluation
                    score = metric(
                        prediction=result.response, actual=example.response
                    )

                    # evaluated benchmark example
                    evaluated_example = BenchmarkEvaluatedExample(
                        score=score,
                        rag_response=result,
                        example=example,
                    )
                    if f is not None:
                        f.write(
                            evaluated_example.model_dump_json_without_embeddings()
                            + "\n"
                        )
                        f.flush()

                    # update running score
                    running_score = self._update_running_score(
                        agg=agg,
                        running_score=running_score,
                        next_score=score,
                        num_examples_seen=num_seen,
                    )

                    num_seen += 1
        finally:
            if f:
                f.close()

        return BenchmarkResult(
            score=running_score,
            metric_name=metric.__class__.__name__,
            num_examples_used=num_seen,
            num_total_examples=benchmark.num_examples,
            evaluations_file=filename.as_posix() if filename else None,
        )



from unittest.mock import MagicMock, patch

import pytest

from fed_rag import NoEncodeRAGSystem, RAGConfig
from fed_rag.base.bridge import BaseBridgeMixin
from fed_rag.base.generator import BaseGenerator
from fed_rag.base.no_encode_knowledge_store import BaseNoEncodeKnowledgeStore
from fed_rag.data_structures import KnowledgeNode, NodeType, SourceNode
from fed_rag.exceptions import RAGSystemError

from .conftest import MockGenerator


class DummyNoEncodeKnowledgeStore(BaseNoEncodeKnowledgeStore):
    nodes: list[KnowledgeNode] = []

    def load_node(self, node: KnowledgeNode) -> None:
        self.nodes.append(node)

    def load_nodes(self, nodes: list[KnowledgeNode]) -> None:
        for n in nodes:
            self.load_node(n)

    def retrieve(
        self, query: str, top_k: int
    ) -> list[tuple[float, KnowledgeNode]]:
        return [(ix, n) for ix, n in enumerate(self.nodes[:top_k])]

    def batch_retrieve(
        self, queries: list[str], top_k: int
    ) -> list[list[tuple[float, KnowledgeNode]]]:
        return [
            [(ix, n) for ix, n in enumerate(self.nodes[:top_k])]
            for jx in range(len(queries))
        ]

    def delete_node(self, node_id: str) -> bool:
        return True

    def clear(self) -> None:
        self.nodes.clear()

    def count(self) -> int:
        return len(self.nodes)

    def persist(self) -> None:
        pass

    def load(self) -> None:
        pass


class DummyNoEncodeNoBatchRetrievalKnowledgeStore(BaseNoEncodeKnowledgeStore):
    nodes: list[KnowledgeNode] = []

    def load_node(self, node: KnowledgeNode) -> None:
        self.nodes.append(node)

    def load_nodes(self, nodes: list[KnowledgeNode]) -> None:
        for n in nodes:
            self.load_node(n)

    def retrieve(
        self, query: str, top_k: int
    ) -> list[tuple[float, KnowledgeNode]]:
        return [(ix, n) for ix, n in enumerate(self.nodes[:top_k])]

    def batch_retrieve(
        self, queries: list[str], top_k: int
    ) -> list[list[tuple[float, KnowledgeNode]]]:
        raise NotImplementedError

    def delete_node(self, node_id: str) -> bool:
        return True

    def clear(self) -> None:
        self.nodes.clear()

    def count(self) -> int:
        return len(self.nodes)

    def persist(self) -> None:
        pass

    def load(self) -> None:
        pass


@pytest.fixture()
def dummy_store() -> BaseNoEncodeKnowledgeStore:
    dummy_store = DummyNoEncodeKnowledgeStore()
    nodes = [
        KnowledgeNode(node_type=NodeType.TEXT, text_content="Dummy text")
        for _ in range(5)
    ]
    dummy_store.load_nodes(nodes)
    return dummy_store


@pytest.fixture()
def dummy_store_no_batch_retrieval() -> BaseNoEncodeKnowledgeStore:
    dummy_store = DummyNoEncodeNoBatchRetrievalKnowledgeStore()
    nodes = [
        KnowledgeNode(node_type=NodeType.TEXT, text_content="Dummy text")
        for _ in range(5)
    ]
    dummy_store.load_nodes(nodes)
    return dummy_store


@pytest.fixture()
def knowledge_store(
    request: pytest.FixtureRequest,
) -> BaseNoEncodeKnowledgeStore:
    return request.getfixturevalue(request.param)


def test_rag_system_init(
    mock_generator: BaseGenerator,
    knowledge_nodes: list[KnowledgeNode],
    dummy_store: BaseNoEncodeKnowledgeStore,
) -> None:
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = NoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=dummy_store,
        rag_config=rag_config,
    )
    assert rag_system.knowledge_store == dummy_store
    assert rag_system.rag_config == rag_config
    assert rag_system.generator == mock_generator


@patch.object(NoEncodeRAGSystem, "generate")
@patch.object(NoEncodeRAGSystem, "_format_context")
@patch.object(NoEncodeRAGSystem, "retrieve")
def test_rag_system_query(
    mock_retrieve: MagicMock,
    mock_format_context: MagicMock,
    mock_generate: MagicMock,
    mock_generator: BaseGenerator,
    knowledge_nodes: list[KnowledgeNode],
    dummy_store: BaseNoEncodeKnowledgeStore,
) -> None:
    # arrange mocks
    source_nodes = [
        SourceNode(score=0.99, node=knowledge_nodes[0]),
        SourceNode(score=0.85, node=knowledge_nodes[1]),
    ]
    mock_retrieve.return_value = source_nodes
    mock_format_context.return_value = "fake context"
    mock_generate.return_value = "fake generation response"

    # build rag system
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = NoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=dummy_store,
        rag_config=rag_config,
    )

    # act
    rag_response = rag_system.query(query="fake query")

    # assert
    mock_retrieve.assert_called_with("fake query")
    mock_format_context.assert_called_with(source_nodes)
    mock_generate.assert_called_with(
        query="fake query", context="fake context"
    )
    assert rag_response.source_nodes == source_nodes
    assert rag_response.response == "fake generation response"
    assert str(rag_response) == "fake generation response"


@patch.object(NoEncodeRAGSystem, "batch_generate")
@patch.object(NoEncodeRAGSystem, "_format_context")
@patch.object(NoEncodeRAGSystem, "batch_retrieve")
def test_rag_system_batch_query(
    mock_batch_retrieve: MagicMock,
    mock_format_context: MagicMock,
    mock_batch_generate: MagicMock,
    mock_generator: BaseGenerator,
    knowledge_nodes: list[KnowledgeNode],
    dummy_store: BaseNoEncodeKnowledgeStore,
) -> None:
    # arrange mocks
    source_nodes = [
        [
            SourceNode(score=0.99, node=knowledge_nodes[0]),
            SourceNode(score=0.85, node=knowledge_nodes[1]),
        ],
        [
            SourceNode(score=0.90, node=knowledge_nodes[0]),
            SourceNode(score=0.80, node=knowledge_nodes[1]),
        ],
    ]
    mock_batch_retrieve.return_value = source_nodes
    mock_format_context.return_value = "fake context"
    mock_batch_generate.return_value = [
        "fake generation response 1",
        "fake generation response 2",
    ]

    # build rag system
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = NoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=dummy_store,
        rag_config=rag_config,
    )
    queries = ["fake query 1", "fake query 2"]

    # act
    rag_responses = rag_system.batch_query(queries=queries)

    # assert
    mock_batch_retrieve.assert_called_with(queries)
    mock_format_context.assert_called_with(source_nodes[1])
    mock_batch_generate.assert_called_with(
        queries, ["fake context", "fake context"]
    )
    assert all(
        res.source_nodes == sn for res, sn in zip(rag_responses, source_nodes)
    )
    assert rag_responses[0].response == "fake generation response 1"
    assert rag_responses[1].response == "fake generation response 2"


@patch.object(MockGenerator, "generate")
def test_rag_system_generate(
    mock_generate: MagicMock,
    mock_generator: MockGenerator,
    dummy_store: BaseNoEncodeKnowledgeStore,
) -> None:
    # arrange mocks
    mock_generate.return_value = "fake generate response"

    # build rag system
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = NoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=dummy_store,
        rag_config=rag_config,
    )

    # act
    res = rag_system.generate(query="fake query", context="fake context")

    # assert
    mock_generate.assert_called_once_with(
        query="fake query", context="fake context"
    )
    assert res == "fake generate response"


@patch.object(MockGenerator, "generate")
def test_rag_system_batch_generate(
    mock_generate: MagicMock,
    mock_generator: MockGenerator,
    dummy_store: BaseNoEncodeKnowledgeStore,
) -> None:
    # arrange mocks
    mock_generate.return_value = [
        "fake generate response 1",
        "fake generate response 2",
    ]

    # build rag system
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = NoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=dummy_store,
        rag_config=rag_config,
    )

    # queries and contexts
    queries = ["fake query 1", "fake query 2"]
    contexts = ["fake context 1", "fake context 2"]

    # act
    res = rag_system.batch_generate(queries=queries, contexts=contexts)

    # assert
    mock_generate.assert_called_once_with(query=queries, context=contexts)
    assert isinstance(res, list)
    assert len(res) == 2
    assert res[0] == "fake generate response 1"
    assert res[1] == "fake generate response 2"


@patch.object(MockGenerator, "generate")
def test_rag_system_batch_generate_list_length_mismatch_error(
    mock_generate: MagicMock,
    mock_generator: MockGenerator,
    dummy_store: BaseNoEncodeKnowledgeStore,
) -> None:
    # build rag system
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = NoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=dummy_store,
        rag_config=rag_config,
    )

    # queries and contexts
    queries = ["fake query 1", "fake query 2"]
    contexts = ["fake context 1"]  # only one context provided

    # act + assert
    with pytest.raises(
        RAGSystemError,
        match="Queries and contexts must have the same length for batch generation.",
    ):
        rag_system.batch_generate(queries=queries, contexts=contexts)

    mock_generate.assert_not_called()


def test_rag_system_format_context(
    mock_generator: MockGenerator,
    dummy_store: BaseNoEncodeKnowledgeStore,
) -> None:
    # build rag system
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = NoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=dummy_store,
        rag_config=rag_config,
    )

    # act
    source_nodes = rag_system.retrieve("mock query")
    formatted_context = rag_system._format_context(source_nodes)

    # assert
    assert formatted_context == "Dummy text\nDummy text"


@pytest.mark.parametrize(
    "knowledge_store",
    ["dummy_store", "dummy_store_no_batch_retrieval"],
    indirect=True,
)
def test_rag_system_batch_retrieve(
    knowledge_store: BaseNoEncodeKnowledgeStore,
    mock_generator: BaseGenerator,
) -> None:
    # build rag system
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = NoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=knowledge_store,
        rag_config=rag_config,
    )

    # queries and expected retrieved source nodes
    queries = ["fake query 1", "fake query 2"]

    # act
    result = rag_system.batch_retrieve(queries)

    # assert
    assert isinstance(result, list)
    assert len(result) == 2
    assert all(isinstance(sn, list) for sn in result)
    assert all(len(sn) == 2 for sn in result)


def test_bridging_no_encode_rag_system(
    mock_generator: MockGenerator,
    dummy_store: BaseNoEncodeKnowledgeStore,
) -> None:
    # arrange

    # create test/mock bridge
    class _TestBridgeMixin(BaseBridgeMixin):
        _bridge_version = "0.1.0"
        _bridge_extra = "my-bridge"
        _framework = "my-bridge-framework"
        _compatible_versions = {"min": "0.1.2", "max": "0.2.0"}
        _method_name = "to_bridge"

        def to_bridge(self) -> None:
            self._validate_framework_installed()
            return None

    class BridgedNoEncodeRAGSystem(_TestBridgeMixin, NoEncodeRAGSystem):
        pass

    # build bridged rag system
    rag_config = RAGConfig(
        top_k=2,
    )
    rag_system = BridgedNoEncodeRAGSystem(
        generator=mock_generator,
        knowledge_store=dummy_store,
        rag_config=rag_config,
    )

    assert _TestBridgeMixin._framework in BridgedNoEncodeRAGSystem.bridges
    assert (
        rag_system.bridges["my-bridge-framework"]
        == _TestBridgeMixin.get_bridge_metadata()
    )

    # cleanup
    del BridgedNoEncodeRAGSystem.bridges[_TestBridgeMixin._framework]
